from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class HeaderEntry:
    header_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "headerName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


@dataclass
class JmespathMappingRecord:
    class Meta:
        name = "JMESPathMappingRecord"

    from_value: Optional[str] = field(
        default=None,
        metadata={
            "name": "from",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    to: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


class ResponseQueryType(Enum):
    JMESPATH_EXPRESSION = "JMESPathExpression"
    XPATH_EXPRESSION = "XPathExpression"
    REGULAR_EXPRESSION = "RegularExpression"
    JMESPATH_MAPPING = "JMESPathMapping"


class RestApiAuthenticationMethod(Enum):
    NO_SECURITY_SCHEME = "NoSecurityScheme"
    BEARER_SECURITY_SCHEME = "BearerSecurityScheme"
    API_KEY_SECURITY_SCHEME = "ApiKeySecurityScheme"
    BASIC_SECURITY_SCHEME = "BasicSecurityScheme"
    DIGEST_SECURITY_SCHEME = "DigestSecurityScheme"
    PSK_SECURITY_SCHEME = "PskSecurityScheme"
    OAUTH1_SECURITY_SCHEME = "OAuth1SecurityScheme"
    OAUTH2_SECURITY_SCHEME = "OAuth2SecurityScheme"
    HAWK_SECURITY_SCHEME = "HawkSecurityScheme"
    AWS_SIGNATURE_SECURITY_SCHEME = "AwsSignatureSecurityScheme"
    NTLM_SECURITY_SCHEME = "NtlmSecurityScheme"
    AKAMAI_EDGE_GRID_SECURITY_SCHEME = "AkamaiEdgeGridSecurityScheme"


@dataclass
class RestApiBasic:
    rest_basic_username: Optional[str] = field(
        default=None,
        metadata={
            "name": "restBasicUsername",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    rest_basic_password: Optional[str] = field(
        default=None,
        metadata={
            "name": "restBasicPassword",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


class RestApiDataType(Enum):
    """
    Rest api specific data types.
    """
    NULL = "null"
    JSON_NUMBER = "JSON_number"
    JSON_STRING = "JSON_string"
    JSON_BOOLEAN = "JSON_boolean"
    JSON_OBJECT = "JSON_object"
    JSON_ARRAY = "JSON_array"


class RestApiInterfaceSelection(Enum):
    """
    Type of Rest Api interface.
    """
    TCPV4 = "TCPV4"
    TCPV6 = "TCPV6"
    URI = "URI"


@dataclass
class HeaderList:
    header: List[HeaderEntry] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class JmespathMapping:
    class Meta:
        name = "JMESPathMapping"

    mapping: List[JmespathMappingRecord] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ResponseQuery:
    query_type: Optional[ResponseQueryType] = field(
        default=None,
        metadata={
            "name": "queryType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    query: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    jmes_path_mappings: Optional[JmespathMapping] = field(
        default=None,
        metadata={
            "name": "jmesPathMappings",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class RestApiServiceCall:
    request_header: Optional[HeaderList] = field(
        default=None,
        metadata={
            "name": "requestHeader",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    request_method: Optional[HttpMethod] = field(
        default=None,
        metadata={
            "name": "requestMethod",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    request_path: Optional[str] = field(
        default=None,
        metadata={
            "name": "requestPath",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    request_body: Optional[str] = field(
        default=None,
        metadata={
            "name": "requestBody",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    response_query: Optional[ResponseQuery] = field(
        default=None,
        metadata={
            "name": "responseQuery",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class RestApiBearer:
    rest_api_service_call: Optional[RestApiServiceCall] = field(
        default=None,
        metadata={
            "name": "restApiServiceCall",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class RestApiDataPointConfiguration:
    """
    Detailed configuration for Rest api data point.
    """
    data_type: Optional[RestApiDataType] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    rest_api_service_call: Optional[RestApiServiceCall] = field(
        default=None,
        metadata={
            "name": "restApiServiceCall",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class RestApiInterfaceDescription:
    """
    Modbus interface properties.
    """
    rest_api_interface_selection: Optional[RestApiInterfaceSelection] = field(
        default=None,
        metadata={
            "name": "restApiInterfaceSelection",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    rest_api_uri: Optional[str] = field(
        default=None,
        metadata={
            "name": "restApiUri",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    rest_api_authentication_method: Optional[RestApiAuthenticationMethod] = field(
        default=None,
        metadata={
            "name": "restApiAuthenticationMethod",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    rest_api_bearer: Optional[RestApiBearer] = field(
        default=None,
        metadata={
            "name": "restApiBearer",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    rest_api_basic: Optional[RestApiBasic] = field(
        default=None,
        metadata={
            "name": "restApiBasic",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )



# Smartgrid Ready Libraries
from sgr_library.modbus_interface import SgrModbusInterface


if __name__ == "__main__":

    # Create interface instance
    interface_file = 'SGr_HeatPump_Test.xml'
    sgr_component = SgrModbusInterface(interface_file)

    # OLD Read Value Method with functional_profile_name and datapoint_name as parameters 
    value = sgr_component.getval('HeatPumpBase', 'HPOpState')
    print(value)

    # NEW Read Value Method with datapoint.
    # First find the datapoint and then get the values from the datapoint object found.
    data_point = sgr_component.find_dp('HeatPumpBase', 'HPOpState')
    value = sgr_component.getval(data_point)
    print(value)

    # We can also search for more info in the datapoint, for example:
    register = sgr_component.get_register_type(data_point)
    multiplicator = sgr_component.get_multiplicator(data_point)
    unit = sgr_component.get_unit(data_point)
    power_10 = sgr_component.get_power_10(data_point)

    # Setval is only implemented in the old style, with functional_profile_name and datapoint_name
    "sgr_component.setval('ActiveEnerBalanceAC', 'ActiveImportAC', 10)"


from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.communicator.communicator_types import CommunicatorBase
from sgrspecification.generic.base_types import GenericAttributeListFunctionalProfile
from sgrspecification.generic.data_point import DataPointBase
from sgrspecification.generic.functional_profile import FunctionalProfileDescription

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class CommunicatorFunctionalProfile(FunctionalProfileDescription):
    """
    Extends the base functional profile type with generic data points.
    """
    generic_attribute_list: Optional[GenericAttributeListFunctionalProfile] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    data_point_list_element: List[DataPointBase] = field(
        default_factory=list,
        metadata={
            "name": "dataPointListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class CommunicatorFrame1(CommunicatorBase):
    """
    Data type definition for a Communicator Description.
    """
    class Meta:
        name = "CommunicatorFrame"

    functional_profile_list_element: List[CommunicatorFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "functionalProfileListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class CommunicatorFrame(CommunicatorFrame1):
    """
    RPT Root Point for Communicator.
    """
    class Meta:
        name = "communicatorFrame"
        namespace = "http://www.smartgridready.com/ns/V0/"



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional
from xsdata.models.datatype import XmlDate
from sgrspecification.generic.base_type_functional_profile_category import FunctionalProfileCategory
from sgrspecification.generic.base_type_level_of_operation_type import LevelOfOperation

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class AlternativeNames:
    """a name list for EEBUS, IEC6850,, SAREF4ENER etc.

    Used to support ontology naming.

    :ivar s_lv1_name: Currently not used. Reserved for the future to
        secure legacy compatibility one we start renaming in future SGr
        label versions
    :ivar work_name: work name for temporary use
    :ivar manuf_name: manufacturers may use an internal wording
    :ivar iec61850_name: IEC 61850 terminology place to add the 61850
        abbreviatuions
    :ivar saref_name: SAREF for ENER terminology place to add the SAREF
        abbreviations (https://saref.etsi.org)
    :ivar eebus_name: EEBUS for terminology place to add the EEBUS
        abbreviations
    :ivar sun_spec_name: sSUNSPEC for terminology place to add the
        www.sunspec.org abbreviations
    :ivar hp_bwp_name: bwp (German Heat Pump Association) for
        terminology place to add the bwp naming for HVAC
    :ivar en17609_name: EN17609 terminology place to add the EU17609
        abbreviations
    """
    s_lv1_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "sLV1Name",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    work_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "workName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    manuf_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "manufName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    iec61850_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "iec61850Name",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    saref_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "sarefName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    eebus_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "eebusName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    sun_spec_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "sunSpecName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    hp_bwp_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "hpBwpName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    en17609_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "en17609Name",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class BitmapEntryFunctionalProfile:
    """
    Maps a device-specific bit mask to a literal.
    """
    literal: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class BitmapEntryProduct:
    """
    Maps a device-specific bit mask to a literal.
    """
    literal: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    hex_mask: Optional[bytes] = field(
        default=None,
        metadata={
            "name": "hexMask",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
            "format": "base16",
        }
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class ChangeLog:
    """
    document history.
    """
    version: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    date: Optional[XmlDate] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    author: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    comment: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


class DataDirectionFunctionalProfile(Enum):
    """
    Read/write direction with read (R), write (W), read-write (RW)
    """
    R = "R"
    W = "W"
    RW = "RW"


class DataDirectionProduct(Enum):
    """
    Read/write direction with read (R), write (W), read-write (RW), read-write
    with persistent storage (RWP), or constant (C)
    """
    C = "C"
    R = "R"
    W = "W"
    RW = "RW"
    RWP = "RWP"


class DeviceCategory(Enum):
    """
    Device Category is inherited from EEBUS, a growing list of supported
    Devices.

    :cvar BATTERY:
    :cvar COMPRESSOR:
    :cvar DEVICE_INFORMATION:
    :cvar DHWCIRCUIT:
    :cvar DHWSTORAGE:
    :cvar DISHWASHER:
    :cvar DRYER:
    :cvar ELECTRICAL_IMMERSION_HEATER:
    :cvar FAN:
    :cvar GAS_HEATING_APPLIANCE:
    :cvar GENERIC:
    :cvar HEATING_BUFFER_STORAGE:
    :cvar HEATING_CIRCUIT:
    :cvar HEATING_OBJECT:
    :cvar HEATING_ZONE:
    :cvar HEAT_PUMP_APPLIANCE:
    :cvar HEAT_SINK_CIRCUIT:
    :cvar HEAT_SOURCE_CIRCUIT:
    :cvar HEAT_SOURCE_UNIT:
    :cvar HVACCONTROLLER:
    :cvar HVACROOM:
    :cvar INSTANT_DHWHEATER:
    :cvar INVERTER:
    :cvar OIL_HEATING_APPLIANCE:
    :cvar PUMP:
    :cvar REFRIGERANT_CIRCUIT:
    :cvar SMART_ENERGY_APPLIANCE:
    :cvar SOLAR_DHWSTORAGE:
    :cvar SOLAR_THERMAL_CIRCUIT:
    :cvar SUB_METER_ELECTRICITY:
    :cvar TEMPERATURE_SENSOR:
    :cvar WASHER:
    :cvar BATTERY_SYSTEM:
    :cvar ELECTRICITY_GENERATION_SYSTEM:
    :cvar ELECTRICITY_STORAGE_SYSTEM:
    :cvar SGCP: EEBUS: GridConnectionPointOfPremises
    :cvar HOUSEHOLD:
    :cvar PVSYSTEM:
    :cvar EV:
    :cvar EVSE:
    :cvar CHARGING_STATION: EEBUS: ChargingOutlet
    :cvar ACTUATOR:
    :cvar CEM:
    """
    BATTERY = "Battery"
    COMPRESSOR = "Compressor"
    DEVICE_INFORMATION = "DeviceInformation"
    DHWCIRCUIT = "DHWCircuit"
    DHWSTORAGE = "DHWStorage"
    DISHWASHER = "Dishwasher"
    DRYER = "Dryer"
    ELECTRICAL_IMMERSION_HEATER = "ElectricalImmersionHeater"
    FAN = "Fan"
    GAS_HEATING_APPLIANCE = "GasHeatingAppliance"
    GENERIC = "Generic"
    HEATING_BUFFER_STORAGE = "HeatingBufferStorage"
    HEATING_CIRCUIT = "HeatingCircuit"
    HEATING_OBJECT = "HeatingObject"
    HEATING_ZONE = "HeatingZone"
    HEAT_PUMP_APPLIANCE = "HeatPumpAppliance"
    HEAT_SINK_CIRCUIT = "HeatSinkCircuit"
    HEAT_SOURCE_CIRCUIT = "HeatSourceCircuit"
    HEAT_SOURCE_UNIT = "HeatSourceUnit"
    HVACCONTROLLER = "HVACController"
    HVACROOM = "HVACRoom"
    INSTANT_DHWHEATER = "InstantDHWHeater"
    INVERTER = "Inverter"
    OIL_HEATING_APPLIANCE = "OilHeatingAppliance"
    PUMP = "Pump"
    REFRIGERANT_CIRCUIT = "RefrigerantCircuit"
    SMART_ENERGY_APPLIANCE = "SmartEnergyAppliance"
    SOLAR_DHWSTORAGE = "SolarDHWStorage"
    SOLAR_THERMAL_CIRCUIT = "SolarThermalCircuit"
    SUB_METER_ELECTRICITY = "SubMeterElectricity"
    TEMPERATURE_SENSOR = "TemperatureSensor"
    WASHER = "Washer"
    BATTERY_SYSTEM = "BatterySystem"
    ELECTRICITY_GENERATION_SYSTEM = "ElectricityGenerationSystem"
    ELECTRICITY_STORAGE_SYSTEM = "ElectricityStorageSystem"
    SGCP = "SGCP"
    HOUSEHOLD = "Household"
    PVSYSTEM = "PVSystem"
    EV = "EV"
    EVSE = "EVSE"
    CHARGING_STATION = "ChargingStation"
    ACTUATOR = "Actuator"
    CEM = "CEM"


@dataclass
class EmptyType:
    pass


class EmptyValue(Enum):
    VALUE = ""


@dataclass
class EnumEntry:
    """
    Maps a device specific ordinal to its literal.
    """
    literal: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    ordinal: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class EnumEntryProductRecord:
    literal: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    ordinal: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class EnumEntryRecordFunctionalProfile:
    literal: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class GenericAttributeFunctionalProfile:
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


class Language(Enum):
    """this is the identification of the language for information to be
    published.

    So far en, de, fr and it is possible
    """
    DE = "de"
    EN = "en"
    FR = "fr"
    IT = "it"


class PowerSource(Enum):
    """
    E0004.
    """
    UNKNOWN = "unknown"
    MAINS1_PHASE = "mains1Phase"
    MAINS3_PHASE = "mains3Phase"
    BATTERY = "battery"
    DC = "dc"


class PresenceLevel(Enum):
    """
    Indicates whether the element is mandatory (M), recommended (R, at least
    one of the R elements must be present), or optional (O, can be omitted)
    """
    M = "M"
    R = "R"
    O = "O"


class ReleaseState(Enum):
    """
    release state of this XML.
    """
    DRAFT = "Draft"
    REVIEW = "Review"
    PUBLISHED = "Published"
    REVOKED = "Revoked"


@dataclass
class ScalingFactor:
    """scaled_value = value * multiplicator * 10^powerof10 This type is used for to
    convert integer datapoint values into floats only"""
    multiplicator: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    powerof10: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


class Units(Enum):
    AMPERES = "AMPERES"
    BARS = "BARS"
    CUBIC_METERS = "CUBIC_METERS"
    CUBIC_METERS_PER_SECOND = "CUBIC_METERS_PER_SECOND"
    DEGREES_CELSIUS = "DEGREES_CELSIUS"
    DEGREES_PHASE = "DEGREES_PHASE"
    HERTZ = "HERTZ"
    HOURS = "HOURS"
    JOULES = "JOULES"
    KILOGRAMS = "KILOGRAMS"
    KILOVOLT_AMPERES = "KILOVOLT_AMPERES"
    KILOVOLT_AMPERE_HOURS = "KILOVOLT_AMPERE_HOURS"
    KILOVOLT_AMPERES_REACTIVE = "KILOVOLT_AMPERES_REACTIVE"
    KILOVOLT_AMPERES_REACTIVE_HOURS = "KILOVOLT_AMPERES_REACTIVE_HOURS"
    KILOWATT_HOURS = "KILOWATT_HOURS"
    KILOWATTS = "KILOWATTS"
    MEGAWATT_HOURS = "MEGAWATT_HOURS"
    METERS = "METERS"
    METERS_PER_SECOND = "METERS_PER_SECOND"
    METERS_PER_SECOND_PER_SECOND = "METERS_PER_SECOND_PER_SECOND"
    MINUTES = "MINUTES"
    NO_UNITS = "NO_UNITS"
    OHMS = "OHMS"
    PARTS_PER_MILLION = "PARTS_PER_MILLION"
    PASCALS = "PASCALS"
    PER_HOUR = "PER_HOUR"
    PERCENT = "PERCENT"
    PERCENT_RELATIVE_HUMIDITY = "PERCENT_RELATIVE_HUMIDITY"
    POWER_FACTOR = "POWER_FACTOR"
    RADIANS = "RADIANS"
    RADIANS_PER_SECOND = "RADIANS_PER_SECOND"
    REVOLUTIONS_PER_MINUTE = "REVOLUTIONS_PER_MINUTE"
    SECONDS = "SECONDS"
    SQUARE_METERS = "SQUARE_METERS"
    VOLT_AMPERES = "VOLT_AMPERES"
    VOLT_AMPERES_REACTIVE = "VOLT_AMPERES_REACTIVE"
    VOLTS = "VOLTS"
    WATT_HOURS = "WATT_HOURS"
    WATTS = "WATTS"
    NONE = "NONE"


@dataclass
class VersionNumber:
    """
    a three digit version mumber system.
    """
    primary_version_number: Optional[int] = field(
        default=None,
        metadata={
            "name": "primaryVersionNumber",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    secondary_version_number: Optional[int] = field(
        default=None,
        metadata={
            "name": "secondaryVersionNumber",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    sub_release_version_number: Optional[int] = field(
        default=None,
        metadata={
            "name": "subReleaseVersionNumber",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class BitmapFunctionalProfile:
    """bitmap for status bits.

    Maps device-specific bit patterns (as hex mask) to literals.
    """
    bitmap_entry: List[BitmapEntryFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "bitmapEntry",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class BitmapProduct:
    """bitmap for status bits.

    Maps device-specific bit patterns (as hex mask) to literals.
    """
    bitmap_entry: List[BitmapEntryProduct] = field(
        default_factory=list,
        metadata={
            "name": "bitmapEntry",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class EnumType:
    """Enum of states.

    Maps device-specific register values (ordinals) to literals.

    :ivar enum_entry:
    :ivar hex_mask: hex mask for bit-wise AND before evaluating the enum
    """
    class Meta:
        name = "Enum"

    enum_entry: List[EnumEntry] = field(
        default_factory=list,
        metadata={
            "name": "enumEntry",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )
    hex_mask: Optional[bytes] = field(
        default=None,
        metadata={
            "name": "hexMask",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "format": "base16",
        }
    )


@dataclass
class EnumMapFunctionalProfile:
    enum_entry: List[EnumEntryRecordFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "enumEntry",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    hex_mask: Optional[bytes] = field(
        default=None,
        metadata={
            "name": "hexMask",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "format": "base16",
        }
    )


@dataclass
class EnumMapProduct:
    enum_entry: List[EnumEntryProductRecord] = field(
        default_factory=list,
        metadata={
            "name": "enumEntry",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )
    hex_mask: Optional[bytes] = field(
        default=None,
        metadata={
            "name": "hexMask",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "format": "base16",
        }
    )


@dataclass
class FunctionalProfileIdentification:
    """
    Specification of design source 0 means: specified by SmartGridready, the
    Profile number follows the SmargGirdready scheme &gt;0 means: specified by
    manufacturer hhhh, the Profile number follows a manufacturers scheme.

    :ivar specification_owner_identification:
    :ivar functional_profile_category: The Profile Identification
        identifies the main profile classes. The enumeration text is
        also documented with numbers being referenced in the profile
        number hhhh.nnnn.uuuu.ss.VV.vv as nnnn
    :ivar functional_profile_type:
    :ivar level_of_operation: levelOfOperation defines a controls
        complexity level m) for read-only monitoring data points level
        1) single contact 2) 2 or more contacts /state controlled
        interface 3) statistical defined characteristics tables 4)
        dynamic real-time control combined with statistical defined
        characteristics tables 5) dynamic real-time control combined
        with dynamic changeable characteristics tables 6) prediction
        based systems
    :ivar version_number:
    """
    specification_owner_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "specificationOwnerIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    functional_profile_category: Optional[FunctionalProfileCategory] = field(
        default=None,
        metadata={
            "name": "functionalProfileCategory",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    functional_profile_type: Optional[str] = field(
        default=None,
        metadata={
            "name": "functionalProfileType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    level_of_operation: Optional[LevelOfOperation] = field(
        default=None,
        metadata={
            "name": "levelOfOperation",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    version_number: Optional[VersionNumber] = field(
        default=None,
        metadata={
            "name": "versionNumber",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class GenericAttributeListFunctionalProfile:
    generic_attribute_list_element: List[GenericAttributeFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "genericAttributeListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class JsonElemFunctionalProfile:
    class Meta:
        name = "JSonElemFunctionalProfile"

    key: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    date: Optional[EmptyValue] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    string: Optional[EmptyValue] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    number: Optional[EmptyValue] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class LegibleDescription:
    """This element us used to extend the definitions with legible text
    elements: a short understandable explanation of the items addressed.

    These elements are used for printed and published information

    :ivar text_element: information to be published
    :ivar language: language identifier de, en, fr, it
    :ivar uri: optional URI pointing towards additional information
    """
    text_element: Optional[str] = field(
        default=None,
        metadata={
            "name": "textElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
            "min_length": 0,
            "max_length": 4000,
        }
    )
    language: Optional[Language] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    uri: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class ReleaseNotes:
    """
    Contains versioning, history and release states.
    """
    state: Optional[ReleaseState] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    remarks: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    change_log: List[ChangeLog] = field(
        default_factory=list,
        metadata={
            "name": "changeLog",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class DataTypeProduct:
    """Generic high-devel data types.

    Implementations of this XSD map these types into native data types
    in their programming language (java, python, go lang, ...)
    """
    enum: Optional[EnumMapProduct] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    bitmap: Optional[BitmapProduct] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    json: Optional[EmptyValue] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    boolean: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int8: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int16: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int32: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int64: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int8_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int8U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int16_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int16U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int32_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int32U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int64_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int64U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    float32: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    float64: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    date_time: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "dateTime",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    string: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class JsonArrayOutputFunctionalProfile:
    class Meta:
        name = "JSonArrayOutputFunctionalProfile"

    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    array: List["JsonArrayOutputFunctionalProfile"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "sequential": True,
        }
    )
    elem: List[JsonElemFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "sequential": True,
        }
    )


@dataclass
class RequestParam:
    param: List[JsonElemFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class GenericAttributeProductEnd:
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    data_type: Optional[DataTypeProduct] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    unit: Optional[Units] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class JsonOutputFunctionalProfile:
    class Meta:
        name = "JSonOutputFunctionalProfile"

    array: List[JsonArrayOutputFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    elem: List[JsonElemFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class DataTypeFunctionalProfile:
    """Generic high-level data types.

    Implementations of this XSD map these types into native data types
    in their programming language (java, python, golang, ...)
    """
    enum: Optional[EnumMapFunctionalProfile] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    bitmap: Optional[BitmapFunctionalProfile] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    json: Optional[JsonOutputFunctionalProfile] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    boolean: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int8: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int16: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int32: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int64: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int8_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int8U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int16_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int16U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int32_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int32U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int64_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int64U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    float32: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    float64: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    date_time: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "dateTime",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    string: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class GenericAttributeListProductEnd:
    generic_attribute_list_element: List[GenericAttributeProductEnd] = field(
        default_factory=list,
        metadata={
            "name": "genericAttributeListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class GenericAttributeProduct:
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    data_type: Optional[DataTypeProduct] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    unit: Optional[Units] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    generic_attribute_list: Optional[GenericAttributeListProductEnd] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class GenericAttributeListProduct:
    generic_attribute_list_element: List[GenericAttributeProduct] = field(
        default_factory=list,
        metadata={
            "name": "genericAttributeListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from sgr_library.api import DataPoint
from sgr_library.api.configuration_parameter import ConfigurationParameter
from sgr_library.api.data_types import DataTypes
from sgr_library.api.function_profile_api import FunctionProfile
from sgrspecification.generic import DeviceCategory, DataDirectionProduct


@dataclass
class DeviceInformation:
    name: str
    manufacture: str
    software_revision: str
    hardware_revision: str
    device_category: DeviceCategory
    is_local: bool


class BaseSGrInterface(ABC):

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def get_function_profiles(self) -> dict[str, FunctionProfile]:
        pass

    @abstractmethod
    def device_information(self) -> DeviceInformation:
        pass

    @abstractmethod
    def configuration_parameter(self) -> list[ConfigurationParameter]:
        pass

    def get_function_profile(self, function_profile_name: str) -> FunctionProfile:
        return self.get_function_profiles()[function_profile_name]

    def get_data_point(self, dp: tuple[str, str]) -> DataPoint:
        return self.get_function_profile(dp[0]).get_data_point(dp[1])

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        data_points = {}
        for fp in self.get_function_profiles().values():
            data_points.update(fp.get_data_points())
        return data_points

    async def read_data(self) -> dict[tuple[str, str], Any]:
        data = {}
        for fp in self.get_function_profiles().values():
            data.update({(fp.name(), key): value for key, value in (await fp.read()).items()})
        return data

    def describe(self) -> tuple[str, dict[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]]:

        data = {}
        for fp in self.get_function_profiles().values():
            key, dps = fp.describe()
            data[key] = dps
        return self.device_information().name, data



from dataclasses import dataclass, field
from enum import Enum
from typing import List

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


class BaudRate(Enum):
    VALUE_300 = "300"
    VALUE_600 = "600"
    VALUE_1200 = "1200"
    VALUE_2400 = "2400"
    VALUE_4800 = "4800"
    VALUE_5600 = "5600"
    VALUE_9600 = "9600"
    VALUE_14400 = "14400"
    VALUE_19200 = "19200"
    VALUE_38400 = "38400"
    VALUE_57600 = "57600"
    VALUE_115200 = "115200"
    VALUE_128000 = "128000"
    VALUE_230400 = "230400"
    VALUE_256000 = "256000"


class ByteLength(Enum):
    VALUE_7 = "7"
    VALUE_8 = "8"


class Parity(Enum):
    EVEN = "EVEN"
    ODD = "ODD"
    NONE = "NONE"


class StopBitLength(Enum):
    VALUE_0 = "0"
    VALUE_1 = "1"
    VALUE_1_5 = "1.5"
    VALUE_2 = "2"


@dataclass
class SerialInterfaceCapability:
    baud_rates_supported: List[BaudRate] = field(
        default_factory=list,
        metadata={
            "name": "baudRatesSupported",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )
    byte_len_supported: List[ByteLength] = field(
        default_factory=list,
        metadata={
            "name": "byteLenSupported",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )
    parity_supported: List[Parity] = field(
        default_factory=list,
        metadata={
            "name": "paritySupported",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )
    stop_bit_len_supported: List[StopBitLength] = field(
        default_factory=list,
        metadata={
            "name": "stopBitLenSupported",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )



from sgrspecification.generic.base_type_functional_profile_category import FunctionalProfileCategory
from sgrspecification.generic.base_type_level_of_operation_type import LevelOfOperation
from sgrspecification.generic.base_types import (
    AlternativeNames,
    BitmapEntryFunctionalProfile,
    BitmapEntryProduct,
    BitmapFunctionalProfile,
    BitmapProduct,
    ChangeLog,
    DataDirectionFunctionalProfile,
    DataDirectionProduct,
    DataTypeFunctionalProfile,
    DataTypeProduct,
    DeviceCategory,
    EmptyType,
    EmptyValue,
    EnumType,
    EnumEntry,
    EnumEntryProductRecord,
    EnumEntryRecordFunctionalProfile,
    EnumMapFunctionalProfile,
    EnumMapProduct,
    FunctionalProfileIdentification,
    GenericAttributeFunctionalProfile,
    GenericAttributeListFunctionalProfile,
    GenericAttributeListProduct,
    GenericAttributeListProductEnd,
    GenericAttributeProduct,
    GenericAttributeProductEnd,
    JsonArrayOutputFunctionalProfile,
    JsonElemFunctionalProfile,
    JsonOutputFunctionalProfile,
    Language,
    LegibleDescription,
    PowerSource,
    PresenceLevel,
    ReleaseNotes,
    ReleaseState,
    RequestParam,
    ScalingFactor,
    Units,
    VersionNumber,
)
from sgrspecification.generic.data_point import (
    DataPointBase,
    DataPointDescription,
)
from sgrspecification.generic.functional_profile import (
    FunctionalProfileBase,
    FunctionalProfileDescription,
)
from sgrspecification.generic.generic_attribute import (
    GenericAttributeFrame,
    GenericAttributeList,
    GenericAttributeListElement,
)
from sgrspecification.generic.sgr_serial_int_capability import (
    BaudRate,
    ByteLength,
    Parity,
    SerialInterfaceCapability,
    StopBitLength,
)

__all__ = [
    "FunctionalProfileCategory",
    "LevelOfOperation",
    "AlternativeNames",
    "BitmapEntryFunctionalProfile",
    "BitmapEntryProduct",
    "BitmapFunctionalProfile",
    "BitmapProduct",
    "ChangeLog",
    "DataDirectionFunctionalProfile",
    "DataDirectionProduct",
    "DataTypeFunctionalProfile",
    "DataTypeProduct",
    "DeviceCategory",
    "EmptyType",
    "EmptyValue",
    "EnumType",
    "EnumEntry",
    "EnumEntryProductRecord",
    "EnumEntryRecordFunctionalProfile",
    "EnumMapFunctionalProfile",
    "EnumMapProduct",
    "FunctionalProfileIdentification",
    "GenericAttributeFunctionalProfile",
    "GenericAttributeListFunctionalProfile",
    "GenericAttributeListProduct",
    "GenericAttributeListProductEnd",
    "GenericAttributeProduct",
    "GenericAttributeProductEnd",
    "JsonArrayOutputFunctionalProfile",
    "JsonElemFunctionalProfile",
    "JsonOutputFunctionalProfile",
    "Language",
    "LegibleDescription",
    "PowerSource",
    "PresenceLevel",
    "ReleaseNotes",
    "ReleaseState",
    "RequestParam",
    "ScalingFactor",
    "Units",
    "VersionNumber",
    "DataPointBase",
    "DataPointDescription",
    "FunctionalProfileBase",
    "FunctionalProfileDescription",
    "GenericAttributeFrame",
    "GenericAttributeList",
    "GenericAttributeListElement",
    "BaudRate",
    "ByteLength",
    "Parity",
    "SerialInterfaceCapability",
    "StopBitLength",
]



from sgr_library.generic_interface import GenericInterface, GenericSGrDeviceBuilder
import asyncio
import os

if __name__ == "__main__":

    async def test_loop():

        print('start loop')

        # We instanciate one interface object with a modbus xml.
        interface_file_modbus = 'xml/abb_terra_01.xml'
        builder = GenericSGrDeviceBuilder()
        modbus_component = builder \
                            .xml_file_path(interface_file_modbus) \
                            .config({}) \
                            .build()


        # We connect to the modbus component. # TODO fix connect thingy
        await modbus_component.connect()

        # We instanciate an interface object with a restapi xml.
        config_file_rest = 'config_CLEMAPEnMon_ressource_default.ini'
        config_file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), config_file_rest)

        # We instanciate a second interface object with a modbusTCP xml.
        interface_file_rest = 'xml/SGr_04_mmmm_dddd_CLEMAPEnergyMonitorEIV0.2.1.xml'
        restapi_component = GenericInterface(interface_file_rest, config_file_path)

        # We authentificate the restapi conneciton
        await restapi_component.authenticate()

        # We create a loop where we request a datapoint with a getval of our restapi 
        # component and a datapoint with a getval of our modbus component.
        while True:
            # instanciate modbus component and use getval to get a value back.
            getval = await modbus_component.getval('CurrentAC', 'CurrentACL1')
            print(getval)

            # instanciate restapi component
            value = await restapi_component.getval('ActivePowerAC', 'ActivePowerACtot')
            print(value)

            await asyncio.sleep(1)

            # you could do the same funciton with a asyncio gather functions if you
            # want to get the variables "concurrently".


    try:
        asyncio.run(test_loop())
    except KeyboardInterrupt:

        # Here we have to close all the sessions...
        # We have to think if we want to open a connection and close it for
        # every getval, or we just leave the user do this.
        print("done")


from sgrspecification.communicator.communicator_frame import (
    CommunicatorFrame1,
    CommunicatorFunctionalProfile,
    CommunicatorFrame,
)
from sgrspecification.communicator.communicator_types import (
    CommunicatorBase,
    CommunicatorCategory,
    CommunicatorInformation,
    CommunicatorTransportService,
    TransportServices,
)

__all__ = [
    "CommunicatorFrame1",
    "CommunicatorFunctionalProfile",
    "CommunicatorFrame",
    "CommunicatorBase",
    "CommunicatorCategory",
    "CommunicatorInformation",
    "CommunicatorTransportService",
    "TransportServices",
]



from enum import Enum

class ResourceError(Enum):
    """
    Default
    """
    NoError = "no_error"
    UnknownError = "unknown_error"

    """
    XML
    """
    XMLParsingError = "xml_parsing_error"
    XMLMissingValueError = "xml_missing_value_error"

    """
    Connection
    """
    ConnectionError = "connection_error"
    CredentialError = "credential_error"
    RequestCreationError = "request_creation_error"
    AuthenticationError = "authentication_error"
    


import configparser
import json
import logging
import ssl
from typing import Any

import aiohttp
import certifi
import jmespath
from aiohttp import ClientResponseError, ClientConnectionError
from cachetools import TTLCache

from generated.product import RestApiAuthenticationMethod
from sgr_library.api import BaseSGrInterface, FunctionProfile, DataPoint, DataPointProtocol, DeviceInformation, \
    ConfigurationParameter
from sgr_library.api.configuration_parameter import build_configurations_parameters
from sgr_library.converters import build_converter
from sgrspecification.generic import DataDirectionProduct
from sgrspecification.product import DeviceFrame
from sgrspecification.product import RestApiFunctionalProfile, RestApiDataPoint
from sgr_library.validators import build_validator

logging.basicConfig(level=logging.ERROR)


def build_rest_data_point(data_point: RestApiDataPoint, function_profile: RestApiFunctionalProfile,
                          interface: 'SgrRestInterface') -> DataPoint:
    protocol = RestDataPoint(data_point, function_profile, interface)
    converter = build_converter(data_point.data_point.unit)
    validator = build_validator(data_point.data_point.data_type)
    return DataPoint(protocol, converter, validator)


class RestDataPoint(DataPointProtocol):

    def __init__(self, rest_api_dp: RestApiDataPoint, rest_api_fp: RestApiFunctionalProfile,
                 interface: 'SgrRestInterface'):
        self._dp = rest_api_dp
        self._fp = rest_api_fp
        self._interface = interface

    def name(self) -> tuple[str, str]:
        return self._fp.functional_profile.functional_profile_name, self._dp.data_point.data_point_name

    async def read(self):
        return await self._interface.getval(self.name()[0], self.name()[1])

    async def write(self, data: Any):
        pass

    def direction(self) -> DataDirectionProduct:
        return self._dp.data_point.data_direction


class RestFunctionProfile(FunctionProfile):

    def __init__(self, rest_api_fp: RestApiFunctionalProfile, interface: 'SgrRestInterface'):
        self._fp = rest_api_fp
        self._interface = interface

        dps = [build_rest_data_point(dp, self._fp, self._interface) for dp in
               self._fp.data_point_list.data_point_list_element]
        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        return self._fp.functional_profile.functional_profile_name

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SgrRestInterface(BaseSGrInterface):
    """
    SmartGrid ready External Interface Class for Rest API
    """

    async def connect(self):
        await self.authenticate()

    def get_function_profiles(self) -> dict[str, FunctionProfile]:
        return self._function_profiles

    def device_information(self) -> DeviceInformation:
        return self._device_information

    def configuration_parameter(self) -> list[ConfigurationParameter]:
        return self._configuration_parameters

    def __init__(self, frame: DeviceFrame, configuration: configparser.ConfigParser):
        # session
        self._device_information = DeviceInformation(
            name=frame.device_name,
            manufacture=frame.manufacturer_name,
            software_revision=frame.device_information.software_revision,
            hardware_revision=frame.device_information.hardware_revision,
            device_category=frame.device_information.device_category,
            is_local=frame.device_information.is_local_control
        )
        self._configuration_parameters = build_configurations_parameters(frame.configuration_list)
        self.ssl_context = ssl.create_default_context(cafile=certifi.where())
        self.connector = aiohttp.TCPConnector(ssl=self.ssl_context)
        self.session = aiohttp.ClientSession(connector=self.connector)
        self.token = None
        self.root = frame
        self._cache = TTLCache(maxsize=100, ttl=5)

        fps = [RestFunctionProfile(profile, self) for profile in
               self.root.interface_list.rest_api_interface.functional_profile_list.functional_profile_list_element]
        self._function_profiles = {fp.name(): fp for fp in fps}
        try:
            description = self.root.interface_list.rest_api_interface.rest_api_interface_description
            request_body = str(description.rest_api_bearer.rest_api_service_call.request_body)
            self.data = json.loads(request_body)
            self.base_url = str(description.rest_api_uri)
            request_path = str(description.rest_api_bearer.rest_api_service_call.request_path)
            self.authentication_url = f'https://{self.base_url}{request_path}'
            logging.info(f"Authentication URL: {self.authentication_url}")

            self.call = self.root.interface_list.rest_api_interface.rest_api_interface_description.rest_api_bearer.rest_api_service_call
            self.headers = {header_entry.header_name: header_entry.value for header_entry in
                            self.call.request_header.header}
        except json.JSONDecodeError:
            logging.exception("Error parsing JSON from the XML file")
            raise
        except configparser.ParsingError:
            logging.exception("Error parsing the configuration file")
            raise
        except configparser.Error:
            logging.exception("General error reading configuration file")
            raise
        except ValueError as e:
            logging.exception(str(e))
            raise
        except Exception as e:
            logging.exception("An unexpected error occurred")
            raise

    async def authenticate(self):
        try:
            async with self.session.post(url=self.authentication_url, headers=self.headers, json=self.data) as res:
                if 200 <= res.status < 300:
                    logging.info(f"Authentication successful: Status {res.status}")
                    try:
                        response = await res.text()
                        token = jmespath.search('accessToken', json.loads(response))
                        if token:
                            self.token = str(token)
                            logging.info("Token retrieved successfully")
                        else:
                            logging.warning("Token not found in the response")
                    except json.JSONDecodeError:
                        logging.error("Failed to decode JSON response")
                    except jmespath.exceptions.JMESPathError:
                        logging.error("Failed to search JSON data using JMESPath")
                else:
                    logging.warning(f"Authentication failed: Status {res.status}")
                    logging.info(f"Response: {await res.text()}")

        except aiohttp.ClientError as e:
            logging.error(f"Network error occurred: {e}")
        except Exception as e:
            logging.error(f"An unexpected error occurred: {e}")

    async def close(self):
        await self.session.close()

    def find_dp(self, root, fp_name: str, dp_name: str):
        """
        Searches the datapoint in the root element.
        :param root: The root element created with the xsdata parser
        :param fp_name: The name of the functional profile in which the datapoint resides
        :param dp_name: The name of the datapoint
        :returns: The datapoint element found in root, if not, returns None.
        """

        if not fp_name or not dp_name:
            raise ValueError("fp_name and dp_name cannot be None or empty")

        try:
            fp = next(
                filter(
                    lambda x: x.functional_profile.functional_profile_name == fp_name,
                    root.interface_list.rest_api_interface.functional_profile_list.functional_profile_list_element
                ),
                None
            )
        except AttributeError as e:
            logging.error(f"Attribute error encountered: {e}")
            return None

        if fp:
            try:
                dp = next(
                    filter(
                        lambda x: x.data_point.data_point_name == dp_name,
                        fp.data_point_list.data_point_list_element
                    ),
                    None
                )
            except AttributeError as e:
                logging.error(f"Attribute error encountered: {e}")
                return None

            if dp:
                return dp
            else:
                logging.error(f"Datapoint '{dp_name}' not found in functional profile '{fp_name}'.")
                return None
        else:
            logging.error(f"Functional profile '{fp_name}' not found.")
            return None

    async def getval(self, fp_name, dp_name):
        try:
            dp = self.find_dp(self.root, fp_name, dp_name)
            if not dp:
                raise ValueError(f"Data point for {fp_name}, {dp_name} not found")

            # Dataclass parsing
            service_call = dp.rest_api_data_point_configuration.rest_api_service_call
            request_path = service_call.request_path

            # Urls string
            url = f'https://{self.base_url}{request_path}'

            query = str(service_call.response_query.query)

            # All headers into dicitonary
            headers = {
                header_entry.header_name: header_entry.value
                for header_entry in service_call.request_header.header
            }
            headers['Authorization'] = f'Bearer {self.token}'

            cache_key = (frozenset(headers), url)

            if cache_key in self._cache:
                response = self._cache.get(cache_key)
            else:
                async with self.session.get(url=url, headers=headers) as res:
                    res.raise_for_status()  # Raises an HTTPError if the HTTP request returned an unsuccessful status code
                    response = await res.json()
                    logging.info(f"Getval Status: {res.status}")

                    response = json.dumps(response)
                    self._cache[cache_key] = response

            value = jmespath.search(query, json.loads(response))
            return value

        except ClientResponseError as e:
            logging.error(f"HTTP error occurred: {e}")
        except ClientConnectionError as e:
            logging.error(f"Connection error occurred: {e}")
        except json.JSONDecodeError as e:
            logging.error(f"Failed to decode JSON: {e}")
        except Exception as e:
            logging.error(f"An unexpected error occurred: {e}")

        return None  # Return None or an appropriate default/fallback value in case of error



__all__ = [
    'BaseSGrInterface',
    'FunctionProfile',
    'DataPoint',
    'DataPointProtocol',
    'DataPointConverter',
    'DataPointValidator',
    'DeviceInformation',
    'ConfigurationParameter',
    'build_configurations_parameters'
]

from sgr_library.api.data_point_api import DataPoint, DataPointProtocol, DataPointConverter, DataPointValidator
from sgr_library.api.device_api import BaseSGrInterface, DeviceInformation
from sgr_library.api.function_profile_api import FunctionProfile
from sgr_library.api.configuration_parameter import ConfigurationParameter, build_configurations_parameters



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional
from sgrspecification.generic.base_types import (
    BitmapProduct,
    EmptyType,
    EnumType,
    ScalingFactor,
)
from sgrspecification.generic.sgr_serial_int_capability import (
    BaudRate,
    ByteLength,
    Parity,
    SerialInterfaceCapability,
    StopBitLength,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


class BitOrder(Enum):
    """
    Modbus bit orders are used to apply different conversion transformations to
    data between Modbus device and IEC61850 devices.
    """
    BIG_ENDIAN = "BigEndian"
    CHANGE_DWORD_ORDER = "ChangeDWordOrder"
    CHANGE_WORD_ORDER = "ChangeWordOrder"
    CHANGE_BYTE_ORDER = "ChangeByteOrder"
    CHANGE_BIT_ORDER = "ChangeBitOrder"


class MasterFunctionsSupported(Enum):
    """Available function/command codes for Master / Clients The various
    reading, writing and other operations are categorized as follows.

    The most "primitive" reads and writes are A number of sources use
    alternative terminology, for example Force Single Coil where the
    standard uses Write Single Coil.[11] Prominent entities within a
    Modbus slave are: ReadDiscreteInputs (code: 2) ReadCoils (code: 1)
    WriteSingleCoil (code: 5) WriteMultipleCoils (code:15)
    ReadInputRegisters (code:4) ReadMultipleHoldingRegisters (code:3)
    WriteSingleHoldingRegister (code:6) WriteMultipleHoldingRegisters
    (code:16) the enum "Primitives" means, that the current register
    Type supports Single Trasnactions If dpSizeNrRegistarts is &gt;1,
    also the multiple access functions must be supported

    :cvar PRIMITIVES: "Primitives" support ReadDiscreteInputs (code: 2)
        ReadCoils (code: 1) WriteSingleCoil (code: 5) ReadInputRegisters
        (code:4) WriteSingleHoldingRegister (code:6) for single Register
        access if numberOfRegisters &gt; 1 "Primitives" supports
        ReadMultipleHoldingRegisters (code:3) WriteMultipleCoils
        (code:15) WriteMultipleHoldingRegisters (code:16) for the
        register Type being adressed
    :cvar READ_DISCRETE_INPUTS: Master function code= 2
    :cvar READ_COILS: Master function code: 1
    :cvar WRITE_SINGLE_COIL: Master function code: 5
    :cvar WRITE_MULTIPLE_COILS: Master function code:15
    :cvar READ_INPUT_REGISTERS: Master function code:4
    :cvar READ_MULTIPLE_HOLDING_REGISTERS: Master function code:3
    :cvar WRITE_SINGLE_HOLDING_REGISTER: Master function code:6
    :cvar WRITE_MULTIPLE_HOLDING_REGISTERS: Master function code:16
    :cvar READ_WRITE_MULTIPLE_REGISTERS: Master function code:23
    :cvar MASK_WRITE_REGISTER: Master function code:22
    :cvar READ_FIFOQUEUE: Master function code:24
    :cvar READ_FILE_RECORD: Master function code:20
    :cvar WRITE_FILE_RECORD: Master function code:21
    :cvar READ_EXCEPTION_STATUS: Master function code:7
    :cvar DIAGNOSTIC: Master function code:8
    :cvar GET_COM_EVENT_COUNTER: Master function code:11
    :cvar GET_COM_EVENT_LOG: Master function code:12
    :cvar REPORT_SLAVE_ID: Master function code:17
    :cvar READ_DEVICE_IDENTIFICATION: Master function code:43
    """
    PRIMITIVES = "Primitives"
    READ_DISCRETE_INPUTS = "ReadDiscreteInputs"
    READ_COILS = "ReadCoils"
    WRITE_SINGLE_COIL = "WriteSingleCoil"
    WRITE_MULTIPLE_COILS = "WriteMultipleCoils"
    READ_INPUT_REGISTERS = "ReadInputRegisters"
    READ_MULTIPLE_HOLDING_REGISTERS = "ReadMultipleHoldingRegisters"
    WRITE_SINGLE_HOLDING_REGISTER = "WriteSingleHoldingRegister"
    WRITE_MULTIPLE_HOLDING_REGISTERS = "WriteMultipleHoldingRegisters"
    READ_WRITE_MULTIPLE_REGISTERS = "ReadWriteMultipleRegisters"
    MASK_WRITE_REGISTER = "MaskWriteRegister"
    READ_FIFOQUEUE = "ReadFIFOQueue"
    READ_FILE_RECORD = "ReadFileRecord"
    WRITE_FILE_RECORD = "WriteFileRecord"
    READ_EXCEPTION_STATUS = "ReadExceptionStatus"
    DIAGNOSTIC = "Diagnostic"
    GET_COM_EVENT_COUNTER = "GetComEventCounter"
    GET_COM_EVENT_LOG = "GetComEventLog"
    REPORT_SLAVE_ID = "ReportSlaveID"
    READ_DEVICE_IDENTIFICATION = "ReadDeviceIdentification"


@dataclass
class ModbusBoolean:
    """
    Modbus specific boolean definition.
    """
    true_value: Optional[int] = field(
        default=None,
        metadata={
            "name": "trueValue",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    false_value: Optional[int] = field(
        default=None,
        metadata={
            "name": "falseValue",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


class ModbusExceptionCode(Enum):
    """
    Type of the Modbus Exceptions sent by Slave (Server) responses.

    :cvar ILLEGAL_FUNCTION: "1: Illegal Function Function code received
        in the query is not recognized or allowed by slave"
    :cvar ILLEGAL_ADDRESS: 2: Illegal Data Address Data address of some
        or all the required entities are not allowed or do not exist in
        slave
    :cvar ILLEGAL_DATA_VALUE: 3: Illegal Data Value Value is not
        accepted by slave
    :cvar SLAVE_FAILURE: 4: Slave (Server) Device Failure Unrecoverable
        error occurred while slave was attempting to perform requested
        action
    :cvar ACK: 5: Acknowledge Slave has accepted request and is
        processing it, but a long duration of time is required. This
        response is returned to prevent a timeout error from occurring
        in the master. Master can next issue a Poll Program Complete
        message to determine whether processing is completed
    :cvar SLAVE_BUSY: 6: Slave (Server) Device Busy Slave is engaged in
        processing a long-duration command. Master should retry later
    :cvar NACK: 7: Negative Acknowledge Slave (Server) cannot perform
        the programming functions. Master should request diagnostic or
        error information from slave
    :cvar MEMORY_PARITY_ERR: 8: Memory Parity Error Slave (Server)
        detected a parity error in memory. Master can retry the request,
        but service may be required on the slave device
    :cvar GTWY_PATH_ERR: 10: Gateway Path Unavailable Specialized for
        Modbus gateways. Indicates a misconfigured gateway
    :cvar GTWY_TARGET_ERR: 11: Gateway Target Device Failed to Respond
        Specialized for Modbus gateways. Sent when slave fails to
        respond
    """
    ILLEGAL_FUNCTION = "IllegalFunction"
    ILLEGAL_ADDRESS = "IllegalAddress"
    ILLEGAL_DATA_VALUE = "IllegalDataValue"
    SLAVE_FAILURE = "SlaveFailure"
    ACK = "ACK"
    SLAVE_BUSY = "SlaveBusy"
    NACK = "NACK"
    MEMORY_PARITY_ERR = "MemoryParityErr"
    GTWY_PATH_ERR = "GtwyPathErr"
    GTWY_TARGET_ERR = "GtwyTargetErr"


class ModbusInterfaceSelection(Enum):
    """
    Type of Modbus interface.
    """
    RTU = "RTU"
    TCPIP = "TCPIP"
    UDPIP = "UDPIP"
    RTU_ASCII = "RTU-ASCII"
    TCPIP_ASCII = "TCPIP-ASCII"
    UDPIP_ASCII = "UDPIP-ASCII"


class ModbusLayer6Deviation(Enum):
    """
    this type is used to manage non standard data type definitions at ISO/OSI
    Layer 6.

    :cvar VALUE_2_REG_BASE1000_L2_H: 2 Registers show a combined value,
        as example in kWh and MWh beginning with lower range @ lower
        address
    :cvar VALUE_2_REG_BASE1000_H2_L: 2 Registers show a combined value,
        as example in kWh and MWh beginning with lower range @ higher
        address
    """
    VALUE_2_REG_BASE1000_L2_H = "2RegBase1000_L2H"
    VALUE_2_REG_BASE1000_H2_L = "2RegBase1000_H2L"


@dataclass
class ModbusTcp:
    """Modbus IP address:Specific P elements for Modbus over IP protocol.

    This definition is partially inherited from 61850-80-5's
    tPTypeModbusIPEnum typedef but reduced to a single interface (not
    redundant adresses RG1 and RG2) and information layer definitions
    """
    port: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    address: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
            "pattern": r"\d+\.\d+\.\d+\.\d+|\{\{.*\}\}",
        }
    )
    slave_id: Optional[int] = field(
        default=None,
        metadata={
            "name": "slaveId",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


class RegisterType(Enum):
    """Type of the Modbus Data.

    For slave-role, definition of the Object Type of the
    data: Discretes Input: Single bit, Read-Only; Coils: Single bit, Read-Write; Input
    Registers: 16-bit word, Read-Only; Holding Registers: 16-bit word, Read-Write.
    """
    COIL = "Coil"
    DISCRETE_INPUT = "DiscreteInput"
    INPUT_REGISTER = "InputRegister"
    HOLD_REGISTER = "HoldRegister"


@dataclass
class AccessProtectionEnabled:
    """Modbus datapoints may be protected by execptions.

    If this is the case, a datapoint may be selected as true with a
    range of supported exceptions. A NOT listed exception means no XY
    exception
    """
    modbus_exception_code: List[ModbusExceptionCode] = field(
        default_factory=list,
        metadata={
            "name": "modbusExceptionCode",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )
    is_enabled: Optional[bool] = field(
        default=None,
        metadata={
            "name": "isEnabled",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class MasterFunctionsSupportedList:
    """Available function/command codes for Master / Clients The various
    reading, writing and other operations are categorized as follows.

    The most "primitive" reads and writes are A number of sources use
    alternative terminology, for example Force Single Coil where the
    standard uses Write Single Coil.[11] Prominent entities within a
    Modbus slave are: ReadDiscreteInputs (code: 2) ReadCoils (code: 1)
    WriteSingleCoil (code: 5) WriteMultipleCoils (code:15)
    ReadInputRegisters (code:4) ReadMultipleHoldingRegisters (code:3)
    WriteSingleHoldingRegister (code:6) WriteMultipleHoldingRegisters
    (code:16) the enum "Primitives" means, that the current register
    Type supports Single Trasnactions If dpSizeNrRegistarts is &gt;1,
    also the multiple access functions must be supported
    """
    master_functions_supported: List[MasterFunctionsSupported] = field(
        default_factory=list,
        metadata={
            "name": "masterFunctionsSupported",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ModbusDataType:
    """
    Modbus specific data types.
    """
    boolean: Optional[ModbusBoolean] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int8: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int16: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int32: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int64: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int8_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int8U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int16_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int16U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int32_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int32U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    int64_u: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "int64U",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    float32: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    float64: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    date_time: Optional[EmptyType] = field(
        default=None,
        metadata={
            "name": "dateTime",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    string: Optional[EmptyType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    enum: Optional[EnumType] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    bitmap: Optional[BitmapProduct] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class ModbusRtu:
    """
    Modbus RTU serial port configuration.
    """
    slave_addr: Optional[int] = field(
        default=None,
        metadata={
            "name": "slaveAddr",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    baud_rate_selected: Optional[BaudRate] = field(
        default=None,
        metadata={
            "name": "baudRateSelected",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    byte_len_selected: Optional[ByteLength] = field(
        default=None,
        metadata={
            "name": "byteLenSelected",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    parity_selected: Optional[Parity] = field(
        default=None,
        metadata={
            "name": "paritySelected",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    stop_bit_len_selected: Optional[StopBitLength] = field(
        default=None,
        metadata={
            "name": "stopBitLenSelected",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    serial_interface_capability: Optional[SerialInterfaceCapability] = field(
        default=None,
        metadata={
            "name": "serialInterfaceCapability",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class TimeSyncBlockNotification:
    """
    Time sync block notifications are used to describe a block of registers
    that can be feched simultaneously.

    :ivar block_cache_identification: Identification (used by data
        points for referencing)
    :ivar first_address: Start address of block
    :ivar size: Block size in number of registers
    :ivar register_type: Modbus register type
    :ivar time_to_live_ms: Block cache time in milliseconds
    """
    block_cache_identification: Optional[str] = field(
        default=None,
        metadata={
            "name": "blockCacheIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    first_address: Optional[int] = field(
        default=None,
        metadata={
            "name": "firstAddress",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    size: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    register_type: Optional[RegisterType] = field(
        default=None,
        metadata={
            "name": "registerType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    time_to_live_ms: Optional[int] = field(
        default=None,
        metadata={
            "name": "timeToLiveMs",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class ModbusAttributes:
    """
    Modbus-specific attributes.

    :ivar scaling_factor: generic value = dataPoint * m * 10^p
    :ivar step_by_increment: each didgit
    :ivar sunssf: a Sunpec specific attribute (scalefactor p -10 ...
        +10) generic value = dataPoint * 10^p note: Sunspec uses sunssf
        usually as Modbus Register with dynamic values check attribute
        "timeAlignedNotification"
    :ivar polling_latency_ms: The time for a master slave communication
        cycle in ms
    :ivar access_protection:
    :ivar layer6_deviation:
    """
    scaling_factor: Optional[ScalingFactor] = field(
        default=None,
        metadata={
            "name": "scalingFactor",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    step_by_increment: Optional[int] = field(
        default=None,
        metadata={
            "name": "stepByIncrement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    sunssf: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    polling_latency_ms: Optional[int] = field(
        default=None,
        metadata={
            "name": "pollingLatencyMs",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    access_protection: Optional[AccessProtectionEnabled] = field(
        default=None,
        metadata={
            "name": "accessProtection",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    layer6_deviation: Optional[ModbusLayer6Deviation] = field(
        default=None,
        metadata={
            "name": "layer6Deviation",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class ModbusDataPointConfiguration:
    """
    Detailed configuration for modbus data point.

    :ivar modbus_data_type:
    :ivar address: 2-bytes-word address: could be a register address (to
        address a modbus register - inputRegister or holdRegister) or a
        part of a bit address (together with bitRank coil -
        discreteInput)
    :ivar bit_rank:
    :ivar register_type:
    :ivar number_of_registers: Number of 16-bit registers of this data
        point
    """
    modbus_data_type: Optional[ModbusDataType] = field(
        default=None,
        metadata={
            "name": "modbusDataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    address: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    bit_rank: Optional[int] = field(
        default=None,
        metadata={
            "name": "bitRank",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_inclusive": 0,
            "max_inclusive": 15,
        }
    )
    register_type: Optional[RegisterType] = field(
        default=None,
        metadata={
            "name": "registerType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    number_of_registers: Optional[int] = field(
        default=None,
        metadata={
            "name": "numberOfRegisters",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class ModbusInterfaceDescription:
    """
    Modbus interface properties.

    :ivar modbus_interface_selection:
    :ivar modbus_tcp:
    :ivar modbus_rtu:
    :ivar first_register_address_is_one: True if the first register
        starts at 1
    :ivar bit_order:
    :ivar master_functions_supported_list:
    """
    modbus_interface_selection: Optional[ModbusInterfaceSelection] = field(
        default=None,
        metadata={
            "name": "modbusInterfaceSelection",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    modbus_tcp: Optional[ModbusTcp] = field(
        default=None,
        metadata={
            "name": "modbusTcp",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    modbus_rtu: Optional[ModbusRtu] = field(
        default=None,
        metadata={
            "name": "modbusRtu",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    first_register_address_is_one: Optional[bool] = field(
        default=None,
        metadata={
            "name": "firstRegisterAddressIsOne",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    bit_order: Optional[BitOrder] = field(
        default=None,
        metadata={
            "name": "bitOrder",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    master_functions_supported_list: Optional[MasterFunctionsSupportedList] = field(
        default=None,
        metadata={
            "name": "masterFunctionsSupportedList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )



from abc import ABC, abstractmethod

from sgr_library.api.data_point_api import DataPoint
from sgr_library.api.data_types import DataTypes
from sgrspecification.generic import DataDirectionProduct


class FunctionProfile(ABC):

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        pass

    def get_data_point(self, dp_name: str) -> DataPoint:
        return self.get_data_points()[(self.name(), dp_name)]

    async def read(self) -> dict[str, DataPoint]:
        return {key[1]: await dp.read() for key, dp in self.get_data_points().items()}

    def describe(self) -> tuple[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]:
        infos = map(lambda dp: dp.describe(), self.get_data_points().values())
        return self.name(), {dp[0][1]: (dp[1], dp[2]) for dp in infos}



from sgr_library import modbus_client, modbus_interface, payload_decoder, restapi_client_async, exceptions, \
    auxiliary_functions, generated

from sgr_library.sgr_device import SGrDevice
__all__ = [
    "modbus_client",
    "modbus_interface",
    "payload_decoder",
    "restapi_client_async",
    "exceptions",
    "auxiliary_functions",
    "generated",
    "SGrDevice"
]



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.base_types import (
    AlternativeNames,
    FunctionalProfileIdentification,
    GenericAttributeListProduct,
    LegibleDescription,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class FunctionalProfileDescription:
    """
    Functional profile properties.

    :ivar functional_profile_name:
    :ivar functional_profile_identification:
    :ivar alternative_names:
    :ivar legible_description: Published and printable information
        related to this functional profile
    :ivar programmer_hints: additional device-specific implementation
        hints for this functional profile
    """
    functional_profile_name: Optional[str] = field(
        default=None,
        metadata={
            "name": "functionalProfileName",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    functional_profile_identification: Optional[FunctionalProfileIdentification] = field(
        default=None,
        metadata={
            "name": "functionalProfileIdentification",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    alternative_names: Optional[AlternativeNames] = field(
        default=None,
        metadata={
            "name": "alternativeNames",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    legible_description: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "legibleDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )
    programmer_hints: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "programmerHints",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "max_occurs": 4,
        }
    )


@dataclass
class FunctionalProfileBase:
    """
    Functional profile element.
    """
    functional_profile: Optional[FunctionalProfileDescription] = field(
        default=None,
        metadata={
            "name": "functionalProfile",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    generic_attribute_list: Optional[GenericAttributeListProduct] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.base_types import (
    DataTypeFunctionalProfile,
    LegibleDescription,
    Units,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class GenericAttributeListElement:
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )
    data_type: Optional[DataTypeFunctionalProfile] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    unit: Optional[Units] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class GenericAttributeList:
    generic_attribute_list_element: List[GenericAttributeListElement] = field(
        default_factory=list,
        metadata={
            "name": "genericAttributeListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class GenericAttributeFrame:
    """
    Generic Attribute.
    """
    class Meta:
        namespace = "http://www.smartgridready.com/ns/V0/"

    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    generic_attribute_list: Optional[GenericAttributeList] = field(
        default=None,
        metadata={
            "name": "genericAttributeList",
            "type": "Element",
        }
    )
    data_type: Optional[DataTypeFunctionalProfile] = field(
        default=None,
        metadata={
            "name": "dataType",
            "type": "Element",
        }
    )
    unit: Optional[Units] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    legible_description: List[LegibleDescription] = field(
        default_factory=list,
        metadata={
            "name": "legibleDescription",
            "type": "Element",
            "max_occurs": 4,
        }
    )



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.data_point import DataPointBase
from sgrspecification.generic.functional_profile import FunctionalProfileBase

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class ContactInterfaceDescription:
    """
    Contact interface properties.

    :ivar number_of_contacts:
    :ivar contact_stabilisation_time_ms: Time in milliseconds until a
        contact has reached a stable state after switching
    """
    number_of_contacts: Optional[int] = field(
        default=None,
        metadata={
            "name": "numberOfContacts",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    contact_stabilisation_time_ms: Optional[int] = field(
        default=None,
        metadata={
            "name": "contactStabilisationTimeMs",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class ContactsDataPointList:
    """
    List of data points.
    """
    data_point_list_element: List[DataPointBase] = field(
        default_factory=list,
        metadata={
            "name": "dataPointListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ContactFunctionalProfile(FunctionalProfileBase):
    data_point_list: List[ContactsDataPointList] = field(
        default_factory=list,
        metadata={
            "name": "dataPointList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ContactFunctionalProfileList:
    """
    List of functional profiles.
    """
    functional_profile_list_element: List[ContactFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "functionalProfileListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class ContactInterface:
    """
    Container for a device with contacts.
    """
    contact_interface_description: Optional[ContactInterfaceDescription] = field(
        default=None,
        metadata={
            "name": "contactInterfaceDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    functional_profile_list: Optional[ContactFunctionalProfileList] = field(
        default=None,
        metadata={
            "name": "functionalProfileList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )



from sgr_library.api import DataPointValidator
from sgrspecification.generic import DataTypeProduct
from sgr_library.validators.validator import IntValidator, EnumValidator, FloatValidator, StringValidator, \
    BooleanValidator


def build_validator(type: DataTypeProduct) -> DataPointValidator:
    if type.int8:
        return IntValidator(8)
    elif type.int16:
        return IntValidator(16)
    elif type.int32:
        return IntValidator(32)
    elif type.int64:
        return IntValidator(64)
    elif type.enum:
        return EnumValidator(type.enum)
    elif type.int8_u:
        return IntValidator(8, signed=False)
    elif type.int16_u:
        return IntValidator(16, signed=False)
    elif type.int32_u:
        return IntValidator(32, signed=False)
    elif type.int64_u:
        return IntValidator(64, signed=False)
    elif type.float32:
        return FloatValidator(32)
    elif type.float64:
        return FloatValidator(64)
    elif type.string:
        return StringValidator()
    elif type.boolean:
        return BooleanValidator()
    elif type.bitmap:
        raise Exception("unsupported validator")  # return BitmapValidator(type.bitmap)
    elif type.date_time:
        raise Exception("unsupported validator")
    raise Exception("unsupported validator")



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Any

from sgr_library.api.data_types import DataTypes
from sgr_library.api.sub_set_units import SubSetUnits
from sgrspecification.generic import DataDirectionProduct

T = TypeVar('T')


class DataPointConverter(ABC, Generic[T]):

    @abstractmethod
    def to_device(self, value: T) -> Any:
        pass

    @abstractmethod
    def from_device(self, value: Any) -> T:
        pass

    @abstractmethod
    def converted_unit(self) -> SubSetUnits:
        pass


class DataPointValidator(ABC):

    @abstractmethod
    def validate(self, value: Any) -> bool:
        pass

    @abstractmethod
    def data_type(self) -> SubSetUnits:
        pass

    def options(self) -> list[str] | None:
        return None


class DataPointProtocol(ABC):

    @abstractmethod
    async def write(self, data: Any):
        pass

    @abstractmethod
    async def read(self) -> Any:
        pass

    @abstractmethod
    def name(self) -> tuple[str, str]:
        pass

    @abstractmethod
    def direction(self) -> DataDirectionProduct:
        pass


class DataPoint(Generic[T]):

    def __init__(self, protocol: DataPointProtocol, converter: DataPointConverter[T], validator: DataPointValidator):
        self._protocol = protocol
        self._converter = converter
        self._validator = validator

    def name(self) -> tuple[str, str]:
        return self._protocol.name()

    async def read(self) -> T:
        value = await self._protocol.read()

        if self._validator.validate(value):
            return self._converter.from_device(value)
        raise Exception(f"invalid value read from device, {value}, validator: {self._validator.data_type()}")

    async def write(self, data: T):
        value = self._converter.to_device(data)
        if self._validator.validate(value):
            await self._protocol.write(value)
        raise Exception("invalid data to write to device")

    def unit(self) -> SubSetUnits:
        return self._converter.converted_unit()

    def direction(self) -> DataDirectionProduct:
        return self._protocol.direction()

    def data_type(self) -> DataTypes:
        return self._validator.data_type()

    def describe(self) -> tuple[tuple[str, str], DataDirectionProduct, DataTypes]:
        return self.name(), self.direction(), self.data_type()

    def options(self) -> list[str]:
        if self._validator.options() == None:
            return []
        return self._validator.options()



# add SmartGridReady libraries
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.context import XmlContext

# Import generated Data Classes
from sgr_library.data_classes.ei_rest_api import SgrRestapideviceDescriptionType
from sgr_library.data_classes.ei_modbus import SgrModbusDeviceDescriptionType

# Smartgrid Ready Libraries
from sgr_library.OLD_restapi_interface import RestapiInterface
from sgr_library.modbus_interface import SgrModbusInterface, find_dp

# os and configparser
import configparser
import os


class smartgridready_component():
    # class for component with smartgridready compatibility

    def __init__(self, bus_type:str, XML_file:str):
        # set XML file with external interface description - OLD MECHANISM   
        #self.interface_file = XML_file
        #self.interface_file_path_default = os.path.join(os.path.dirname(os.path.realpath(__file__)),  self.interface_file)
        
        # instanciate an XML parser - OLD MECHANISM
        #self.Parser = Parser(self.interface_file_path_default) 
        #self.address = self.Parser.address_return()             # get Modbus address and port
        #self.port = self.Parser.port_return()

        # instanciate an external interface (EI) to connect with the device - OLD MECHANISM
        #if bus_type == 'modbus':
        #    self.EI = EI4Modbus(self.port, self.address, self.interface_file)
        #else:    
        #    self.EI = None                  # --- add additional interfaces here
        
        # read configuration and interface files - NEW MECHANISM
        config_file_path_default = os.path.join( os.path.dirname(os.path.realpath(__file__)), config_file)
        config_ressource = configparser.ConfigParser()
        config_ressource.read(config_file_path_default)
        
        interface_file = XML_file
        parser = XmlParser(context=XmlContext())
        if bus_type == 'modbus':
            self.root = parser.parse(interface_file, SgrModbusDeviceDescriptionType)   # ADDED # NOT NEEDED
            self.sgr_reader_interface = SGrModbusInterface(interface_file) # ADDED
        elif bus_type == 'rest_api':
            self.root = parser.parse(interface_file, SgrRestapideviceDescriptionType)
            self.sgr_reader_interface = RestapiInterface(interface_file, config_ressource)
        else:
            self.root = None
            self.sgr_reader_interface = None     

    def read_value(self, functional_profile: str, data_point: str):
        # read one value from a given data point within a functional profile
        
        value = 0
        unit = None
        error_code = 100         # not found

        """fp_profiles = self.Parser.fp_return()   # get all functional profiles in XML file - OLD MECHANISM
        fp_profiles = self.root.fp_list_element # get all functional profiles in XML file - NEW MECHANISM
        for fp in fp_profiles:
            if fp == functional_profile:
                for dp in fp_profiles[fp]:
                    if dp == data_point:
                        name = dp.data_point[0].datapoint_name                     # get name - NEW MECHANISM
                        #value = EI.value_decoder(int(fp_profiles[fp][dp]['addr']))  # get value - OLD MECHANISM
                        value = dp.data_point[0].datapoint_value                    # get value - NEW MECHANISM, IS SYNTAX CORRECT ???
                        #Datapoint attribute: dp.data_point[0].basic_data_type.float32  # get attribute - WHAT'S THE DIFFERENCE BETWEEN ATTRIBUTE AND VALUE?
                        #unit = fp_profiles[fp][dp]['unit']                         # get unit - OLD MECHANISM
                        unit = dp.data_point[0].datapoint_unit                      # get unit - NEW MECHANISM, IS SYNTAX CORRECT, DO UNITS EXIST ???
                        error_code = 0"""

        # This makes all the search for you and gets you the decoded value.
        value = self.sgr_reader_interface.get_val(functional_profile, data_point)
        # If you want unit and name you can use the datapoint_info function. This one has to be refactored,
        # so it is more intuitive for the user.
        # [0] for addres, [1] for size, [2] for bitrank, [3] for register_type, [4] for unit, [5] for multiplicator, [6] for power_off
        
        unit = self.sgr_reader_interface.datapoint_info[4]

        # The datapoint name I did not add yet to the datapoint_info function but you can manually find it doing this for example:

        dp = find_dp(self.root, functional_profile, data_point)
        name = dp.data_point[0].datapoint_name
        error_code = 0
        
        return [value, unit, error_code]
                
    def write_value(self, functional_profile:str, data_point:str, value:float):
        # write one value to a given data point within a functional profile

        self.sgr_reader_interface.setval(functional_profile, data_point, value)

        None    # --- code under contruction, EI4_ModbusTCP does not support writing yet ---
        error_code = 0

        return error_code


from typing import Any

from sgr_library.api import DataPointValidator
from sgr_library.api.data_types import DataTypes
from sgrspecification.generic import EnumMapProduct


class UnsupportedValidator(DataPointValidator):
    def __init__(self, type: Any):
        self._type = type

    def validate(self, value: Any) -> bool:
        return False


class EnumValidator(DataPointValidator):
    def __init__(self, type: EnumMapProduct):
        self._type = type
        literals = {entry.literal for entry in type.enum_entry}
        ordinals = {entry.ordinal for entry in type.enum_entry}
        literals.union(ordinals)
        literals.discard(None)
        self._valid_entries: list[str] = literals

    def validate(self, value: Any) -> bool:
        return value in self._valid_entries

    def data_type(self):
        return DataTypes.ENUM

    def options(self) -> list[str] | None:
        return self._valid_entries


class IntValidator(DataPointValidator):

    def __init__(self, size: int, signed: bool = False):
        self._size = size
        if signed:
            self._lower_bound = -2 ** (self._size - 1)
            self._upper_bound = 2 ** (self._size - 1)
        else:
            self._lower_bound = 0
            self._upper_bound = 2 ** self._size - 1

    def validate(self, value: Any) -> bool:
        print(isinstance(value, int), self._lower_bound <= value <= self._upper_bound)
        return isinstance(value, int) and self._lower_bound <= value <= self._upper_bound

    def data_type(self):
        return DataTypes.INT


class FloatValidator(DataPointValidator):

    def __init__(self, size: int):
        self._size = size

    def validate(self, value: Any) -> bool:
        return isinstance(value, float) or isinstance(value,
                                                      int)  # added int here as the test sensor returns int and i have nothing else to test

    def data_type(self):
        return DataTypes.FLOAT


class StringValidator(DataPointValidator):

    def validate(self, value: Any) -> bool:
        return isinstance(value, str)

    def data_type(self):
        return DataTypes.STRING


class BooleanValidator(DataPointValidator):

    def validate(self, value: Any) -> bool:
        return isinstance(value, bool)

    def data_type(self):
        return DataTypes.BOOLEAN



"""SGr Python Exceptions.

Custom exceptions to be used in the SGr Library code.
"""

class MyCustomException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__()

    def __str__(self):
        """Return string representation."""
        return f"Custom Error: {self.message}"

class XMLValueException(Exception): ...
class ParsingError(Exception): ...

# Search DP exceptions
class FunctionalProfileException(ValueError): ...
class DataPointException(ValueError): ...
class DataProcessingError(Exception): ...
class DeviceInformationError(Exception): ...

# Modbus exceptions
class RegisterError(Exception): ...
class InvalidEndianType(Exception): ...


from enum import Enum

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


class LevelOfOperation(Enum):
    """
    maturity level concerning SmartGridread functionality.

    :cvar M: m: monitoring only
    :cvar VALUE_1: 1: binary writable data points
    :cvar VALUE_2: 2: discrete writable data points
    :cvar VALUE_3: 3: fixed set of characteristic curves
    :cvar VALUE_4: 4: dynamic set values
    :cvar VALUE_5: 5: dynamic characteristic curves
    :cvar VALUE_6: 6: predictive
    :cvar VALUE_1M: 1m: binary writable data points plus monitoring
    :cvar VALUE_2M: 2m: discrete writable data points plus monitoring
    :cvar VALUE_3M: 3m: fixed set of characteristic curves plus
        monitoring
    :cvar VALUE_4M: 4m: dynamic set values plus monitoring
    :cvar VALUE_5M: 5m: dynamic characteristic curves plus monitoring
    :cvar VALUE_6M: 6m: predictive plus monitoring
    """
    M = "m"
    VALUE_1 = "1"
    VALUE_2 = "2"
    VALUE_3 = "3"
    VALUE_4 = "4"
    VALUE_5 = "5"
    VALUE_6 = "6"
    VALUE_1M = "1m"
    VALUE_2M = "2m"
    VALUE_3M = "3m"
    VALUE_4M = "4m"
    VALUE_5M = "5m"
    VALUE_6M = "6m"



from setuptools import setup, find_packages

setup(
    name='SGrSpecificationPythontks4r',
    version='0.1.7',
    packages=find_packages(),
    url='https://github.com/robin-schoch/SGrPython',
    license='',
    author='robin',
    author_email='',
    description='',
    long_description='',
    long_description_content_type='text/plain',
)



from typing import Any

from sgr_library.api import DataPointConverter
from sgr_library.api.data_point_api import T
from sgr_library.api.sub_set_units import SubSetUnits
from sgrspecification.generic import Units


class VoltDataPointConverter(DataPointConverter[int]):

    def __init__(self, unit: Units):
        self._unit = unit
        self._factor = 1
        match unit:
            case unit.VOLTS:
                self._factor = 1

    def to_device(self, value: int) -> Any:
        return value / self._factor

    def from_device(self, value: Any) -> int:
        return value * self._factor

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.VOLTS

class ReactiveEnergyConverter(DataPointConverter[int]):

    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        pass


class PowerOverTimeConverter(DataPointConverter[int]):
    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        pass


class CurrentConverter(DataPointConverter[int]):

    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        pass


class PowerConverter(DataPointConverter[int]):

    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        pass

class TemperatureConverter(DataPointConverter[int]):

    def __init__(self, unit: Units) -> None:
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.DEGREES_CELSIUS

class HoursConverter(DataPointConverter[int]):

    def __init__(self, unit: Units) -> None:
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.HOURS

class MinutesConverter(DataPointConverter[int]):

    def __init__(self, unit: Units) -> None:
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.MINUTES

class SecondsConverter(DataPointConverter[int]):

    def __init__(self, unit: Units) -> None:
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.SECONDS

class PercentConverter(DataPointConverter[int]):

    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return value

    def from_device(self, value: Any) -> T:
        return value

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.PERCENT


class NoneConverter(DataPointConverter[None]):

    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: T) -> Any:
        return None

    def from_device(self, value: Any) -> None:
        return None

    def converted_unit(self) -> SubSetUnits:
        return SubSetUnits.NONE

class UnsupportedConverter(DataPointConverter[Any]):

    def __init__(self, unit: Units):
        self._unit = unit

    def to_device(self, value: Any) -> Any:
        raise Exception(f"unsupported unit, {self._unit}")

    def from_device(self, value: Any) -> Any:
        raise Exception(f"unsupported unit, {self._unit}")

    def converted_unit(self) -> SubSetUnits:
        raise Exception(f"unsupported unit, {self._unit}")



# Diese Python file basiert dem "smartGrid_Ready_python_restapi_alpha_read_energy_monitor.py" von Gino (Merci!!)

import logging
import threading
import time
import configparser
import os
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.context import XmlContext

# Import generated Data Classes
from sgr_library.data_classes.ei_rest_api import SgrRestapideviceDescriptionType
from sgr_library.data_classes.ei_modbus import SgrModbusDeviceDescriptionType

# Smartgrid Ready Libraries
from sgr_library.OLD_restapi_interface import RestapiInterface


###################################################################################
# simple reader
###################################################################################
def Reader(sgr_interface, stop_reader):


    interface = sgr_interface
    logging.info("Reader: started")
    
    # Private_config and communication_channel can be deleted once we add this to the python library
    interface.get_val_detailed("ActivePowerAC", "ActivePowerACtot") 
    
    while not stop_reader.is_set():
        # Repeat loop every 4 sec
        time.sleep(2)
        
        cycle_start_timestamp = time.time()
        interface.cycle_start_timestamp = cycle_start_timestamp
        logging.info("Reader: cycle started at {}".format(time.ctime(cycle_start_timestamp)))
        # We loop inside the functional profiles and then into the datapoints
        for fp in root.fp_list_element:
            for dp in fp.dp_list_element:
                FP_Name = fp.functional_profile.profile_name
                DP_Name = dp.data_point[0].datapoint_name
                interface.get_val_detailed(FP_Name, DP_Name)
            
        time.sleep(2)

    logging.info("Reader: received stop_reader. Exiting")


###################################################################################
# Simulationsparameter
###################################################################################
simulation_duration = 17 #3  # stop the simulation after X seconds
config_file = 'config_CLEMAPEnMon_ressource_default.ini'
#interface_file = 'CLEMAP_ModbusMeterV0.1.1.xml'
interface_file = 'SGr_04_0018_CLEMAP_EIcloudEnergyMonitorV0.2.1.xml'

###################################################################################
# Main Simulationsprogramm
###################################################################################
if __name__ == "__main__":
    format = "%(asctime)s: %(message)s"
    logging.basicConfig( format=format, level=logging.INFO, datefmt="%H:%M:%S")

    # read configuration and interface files
    config_file_path_default = os.path.join( os.path.dirname(os.path.realpath(__file__)), config_file)
    config_ressource = configparser.ConfigParser()
    config_ressource.read(config_file_path_default)
    
    interface_file = 'SGr_04_0018_CLEMAP_EIcloudEnergyMonitorV0.2.1.xml'
    parser = XmlParser(context=XmlContext())
    root = parser.parse(interface_file, SgrRestapideviceDescriptionType)


    sgr_reader_interface = RestapiInterface(interface_file, config_ressource)


    # print interface
    print('***** SGr FP *****')
    print(f"device {root.device_name} of kind {root.device_profile.device_kind} by manufacturer {root.manufacturer_name}")

    for fp in root.fp_list_element:
        FP_Name = fp.functional_profile.profile_name
        for dp in fp.dp_list_element:
            DP_Name = dp.data_point[0].datapoint_name
            print(f' SGr DP {DP_Name}')
            print(f'    Datapoint Attribute: {dp.data_point[0].basic_data_type.float32}')
    print('*****')



    # loop through a number of iterations
    reader_stop_event = threading.Event()
    logging.info("Main: starting reader ...")

    reader_thread = threading.Thread(target = Reader, args =(sgr_reader_interface, reader_stop_event, ))
    reader_thread.start()

    simulation_start_timestamp = time.time()

    while time.time() - simulation_start_timestamp < simulation_duration:
        logging.info(".")
        time.sleep(1)

    logging.info("Main: stopping reader ...")
    reader_stop_event.set()
    time.sleep(2)
    logging.info("Main: Bye Bye!")



from dataclasses import dataclass, field
from typing import List, Optional
from sgrspecification.generic.data_point import DataPointBase
from sgrspecification.generic.functional_profile import FunctionalProfileBase
from sgrspecification.product.rest_api_types import (
    RestApiDataPointConfiguration,
    RestApiInterfaceDescription,
)

__NAMESPACE__ = "http://www.smartgridready.com/ns/V0/"


@dataclass
class RestApiDataPoint(DataPointBase):
    rest_api_data_point_configuration: Optional[RestApiDataPointConfiguration] = field(
        default=None,
        metadata={
            "name": "restApiDataPointConfiguration",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
        }
    )


@dataclass
class RestApiDataPointList:
    """
    List of data points.
    """
    data_point_list_element: List[RestApiDataPoint] = field(
        default_factory=list,
        metadata={
            "name": "dataPointListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class RestApiFunctionalProfile(FunctionalProfileBase):
    data_point_list: Optional[RestApiDataPointList] = field(
        default=None,
        metadata={
            "name": "dataPointList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )


@dataclass
class RestApiFunctionalProfileList:
    """
    List of functional profiles.
    """
    functional_profile_list_element: List[RestApiFunctionalProfile] = field(
        default_factory=list,
        metadata={
            "name": "functionalProfileListElement",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "min_occurs": 1,
        }
    )


@dataclass
class RestApiInterface:
    """
    Container for a rest api device.
    """
    rest_api_interface_description: Optional[RestApiInterfaceDescription] = field(
        default=None,
        metadata={
            "name": "restApiInterfaceDescription",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )
    functional_profile_list: Optional[RestApiFunctionalProfileList] = field(
        default=None,
        metadata={
            "name": "functionalProfileList",
            "type": "Element",
            "namespace": "http://www.smartgridready.com/ns/V0/",
            "required": True,
        }
    )



import logging
from dataclasses import dataclass
from typing import Optional, Tuple, Dict, Any, Iterable
from pymodbus.constants import Endian
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.context import XmlContext
import time

from sgr_library.api import build_configurations_parameters
#from sgr_library.data_classes.ei_modbus import SgrModbusDeviceDescriptionType
from sgrspecification.product import DeviceFrame
#from sgr_library.data_classes.ei_modbus.sgr_modbus_eidevice_frame import SgrModbusDataPointsFrameType
from sgr_library.modbusRTU_client import SGrModbusRTUClient
from sgr_library.auxiliary_functions import get_port, get_endian, find_dp, get_baudrate, get_slave_rtu, get_parity



class SgrModbusRtuInterface:
    # a global Modbus client
    globalModbusRTUClient = None

    def __init__(self, frame: DeviceFrame) -> None:
        """
        Creates a connection from xml file data.
        Parses the xml file with xsdata library.
        :param xml_file: Name of the xml file to parse
        """

        self.root = frame
        #self.root = parser.parse(interface_file, SgrModbusDeviceDescriptionType)
        self._configuration_params = build_configurations_parameters(frame.configuration_list)
        self.port = get_port(self.root) #TODO überlegungen machen wo Port untergebracht wird
        self.baudrate = get_baudrate(self.root)
        self.parity = get_parity(self.root)
        self.slave_id = get_slave_rtu(self.root)
        self.byte_order = get_endian(self.root)

        # check if there already exists a ModbusRTUClient for the communication over ModbusRTU
        if SgrModbusRtuInterface.globalModbusRTUClient is None:
            self.client = SGrModbusRTUClient(str(self.port), str(self.parity), int(self.baudrate))
            SgrModbusRtuInterface.globalModbusRTUClient = self.client
        else:
            self.client = SgrModbusRtuInterface.globalModbusRTUClient



    def get_pymodbus_client(self):
        """
        returns the pymodbus client object
        """
        return self.client.client

    #TODO
    def get_dp_attribute(self, datapoint: str, attribute: str):
        """
        Searches for a specific attribute in the datapoint via a key.
        :param attribute"address", "size", "bitrank", "data_type", "register_type", "unit", "multiplicator", "power_of", "name"
        :returns: The chosen attribute.
        """
        #TODO
        ...

    #TODO assign multiple dispatch to the function.
    '''def getval(self, fp_name: str, dp_name: str) -> float:
        """
        Reads datapoint value.
        :param fp_name: The name of the funcitonal profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.
        :returns: The current decoded value in the datapoint register.
        """
        dp = find_dp(self.root, fp_name, dp_name)
        address = self.get_address(dp)
        size = self.get_size(dp)
        data_type = self.get_datatype(dp)
        reg_type = self.get_register_type(dp)
        return self.client.value_decoder(address, size, data_type, reg_type)'''

    # getval with multiple dispatching
    def getval(self, *parameter) -> float:
        """
        Reads datapoint value.

        :dp: The already obtained datapoint object

        2 parameters alternative:
        :param fp_name: The name of the functional profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.

        :returns: The current decoded value in the datapoint register.
        """
        if len(parameter) == 2:
            dp = find_dp(self.root, parameter[0], parameter[1])
        else:
            dp = parameter[0]
        address = self.get_address(dp)
        size = self.get_size(dp)
        data_type = self.get_datatype(dp)
        reg_type = self.get_register_type(dp)
        slave_id = self.slave_id
        order = self.byte_order

        logging.debug(f"getVal() with address: {address}, size: {size}, data_type:{data_type}, slave_id: {slave_id}, order: {order}") # for debugging
        return self.client.value_decoder(address, size, data_type, reg_type, slave_id, order)

    def setval(self, fp_name: str, dp_name: str, value: float) -> None:
        """
        Writes datapoint value.
        :param fp_name: The name of the funcitonal profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.
        :param value: The value that is to be written on the datapoint.
        """
        dp = find_dp(self.root, fp_name, dp_name)
        address = self.get_address(dp)
        data_type = self.get_datatype(dp)
        slave_id = self.slave_id
        order = self.byte_order
        self.client.value_encoder(address, value, data_type, slave_id, order)

    def get_device_profile(self):
        print(f"Brand Name: {self.root.device_profile.brand_name}")
        print(f"Nominal Power: {self.root.device_profile.nominal_power}")
        print(f"Level of Operation: {self.root.device_profile.dev_levelof_operation}")
        return (self.root.device_profile)

    def get_register_type(self, dp) -> str:
        """
        Returns register type E.g. "HoldRegister"
        :param fp_name: The name of the functional profile
        :param dp_name: The name of the data point.
        :returns: The type of the register
        """
        register_type = dp.modbus_data_point[0].modbus_first_register_reference.register_type.value
        return register_type

    def get_datatype(self, dp) -> str:
        datatype = dp.modbus_data_point[0].modbus_data_type.__dict__
        for key in datatype:
            if datatype[key] != None:
                return key
        print('data_type not available')

    def get_bit_rank(self, dp):
        bitrank = dp.modbus_data_point[0].modbus_first_register_reference.bit_rank
        return bitrank

    def get_address(self, dp):
        address = dp.modbus_data_point[0].modbus_first_register_reference.addr
        return address


    def get_size(self, dp):
        size = dp.modbus_data_point[0].dp_size_nr_registers
        return size

    def get_multiplicator(self, dp)->int:
        multiplicator = dp.modbus_attr[0].scaling_by_mul_pwr.multiplicator
        return multiplicator

    def get_power_10(self, dp)->int:
        power_10 = dp.modbus_attr[0].scaling_by_mul_pwr.powerof10
        return power_10

    def get_unit(self, dp):
        unit = dp.data_point.unit.value
        return unit

    def get_name(self, dp):
        name = dp.data_point[0].datapoint_name
        return name

    def find_dp(self, fp_name: str, dp_name: str):
        """
        Searches the datapoint in the root element.
        :param root: The root element created with the xsdata parser
        :param fp_name: The name of the funcitonal profile in which the datapoint resides
        :param dp_name: The name of the datapoint
        :returns: The datapoint element found in root, if not, returns None.
        """
        for fp in self.root.fp_list_element:
                if fp_name == fp.functional_profile.profile_name:
                    #Secondly we filter the datpoint name
                    for dp in fp.dp_list_element:
                        if dp_name == dp.data_point[0].datapoint_name:
                            return dp
        return None

    def get_device_name(self)-> str:
        """
        returns the s_LV1_Name of the device
        """
        return self.root.device_profile.dev_name_list.s_lv1_name

    def get_modbusInterfaceSelection(self) -> str:
        """
        Returns selected Modbus Interface in XML file
        """
        return (self.root.modbus_interface_desc.modbus_interface_selection.value)

    def get_manufacturer(self):
        """
        returns the manufacturer name of the device
        """
        return self.root.device_profile.dev_name_list.s_manuf_name

    def set_slave_id(self,slave_id: int):
        """
        changes the slave id for the instance
        """
        self.slave_id = slave_id

    # TODO a getval for L1, L2 and L3 at the same time

if __name__ == "__main__":
    starting_time = time.time()
    print('start')
    interface_file = '../xml_files/SGr_04_0016_xxxx_ABBMeterV0.2.1.xml'
    a = SgrModbusRtuInterface(interface_file)
    print('ActivePowerACtot :',a.getval('ActivePowerAC', 'ActivePowerACtot'))
    # Power = a.client.value_decoder(0x5B14,2,"int32","HoldingRegister",1,Endian.Big)
    # print(str(Power*0.01) + "W")


    #a.setval('ActiveEnerBalanceAC', 'ActiveImportAC', 9000)

    dp = find_dp(a.root, 'ActiveEnerBalanceAC', 'ActiveImportAC')
    print("ActiveImportAC :",a.getval(dp))

    a.client.client.close()
    print("print finished")



import re

from sgr_library.api import BaseSGrInterface, FunctionProfile, DeviceInformation, ConfigurationParameter
from sgr_library.generic_interface import GenericSGrDeviceBuilder


class SGrDevice(BaseSGrInterface):

    def __init__(self):
        self._builder: GenericSGrDeviceBuilder = GenericSGrDeviceBuilder()
        self._interface: BaseSGrInterface | None = None
        self._configuration_params: list[str] = []

    def is_connect(self):
        return self._interface is not None

    async def connect(self):
        self._interface = self._builder.build()
        await self._interface.connect()

    def update_config(self, config: dict | str) -> 'SGrDevice':
        if isinstance(config, str):
            self._builder.config_file_path(config)
        else:
            self._builder.config(config)

        return self

    def show_replacement(self):
        content, _ = self._builder.replace_variables()
        return content

    def update_xml_spec(self, spec: str) -> 'SGrDevice':
        content = spec
        if spec.startswith("<?xml"):
            self._builder.xml_string(spec)
        else:
            self._builder.xml_file_path(spec)
            content = self._builder.get_spec_content()
        return self

    def get_function_profiles(self) -> dict[str, FunctionProfile]:
        return self._interface.get_function_profiles()

    def build(self):
        self._interface = self._builder.build()

    def device_information(self) -> DeviceInformation:
        return self._interface.device_information()

    def configuration_parameter(self) -> list[ConfigurationParameter]:
        return self._interface.configuration_parameter()

    def get_frame(self):
        return self._interface.root



from sgrspecification.product.contact_interface import (
    ContactFunctionalProfile,
    ContactFunctionalProfileList,
    ContactInterface,
    ContactInterfaceDescription,
    ContactsDataPointList,
)
from sgrspecification.product.generic_interface import (
    GenericDataPointList,
    GenericFunctionalProfile,
    GenericFunctionalProfileList,
    GenericInterface,
)
from sgrspecification.product.modbus_interface import (
    ModbusDataPoint,
    ModbusDataPointList,
    ModbusFunctionalProfile,
    ModbusFunctionalProfileList,
    ModbusInterface,
)
from sgrspecification.product.modbus_types import (
    AccessProtectionEnabled,
    BitOrder,
    MasterFunctionsSupported,
    MasterFunctionsSupportedList,
    ModbusAttributes,
    ModbusBoolean,
    ModbusDataPointConfiguration,
    ModbusDataType,
    ModbusExceptionCode,
    ModbusInterfaceDescription,
    ModbusInterfaceSelection,
    ModbusLayer6Deviation,
    ModbusRtu,
    ModbusTcp,
    RegisterType,
    TimeSyncBlockNotification,
)
from sgrspecification.product.product import (
    ConfigurationDescription,
    ConfigurationList,
    ConfigurationListElement,
    DeviceFrame,
    DeviceInformation,
    InterfaceList,
)
from sgrspecification.product.rest_api_interface import (
    RestApiDataPoint,
    RestApiDataPointList,
    RestApiFunctionalProfile,
    RestApiFunctionalProfileList,
    RestApiInterface,
)
from sgrspecification.product.rest_api_types import (
    HeaderEntry,
    HeaderList,
    HttpMethod,
    JmespathMapping,
    JmespathMappingRecord,
    ResponseQuery,
    ResponseQueryType,
    RestApiAuthenticationMethod,
    RestApiBasic,
    RestApiBearer,
    RestApiDataPointConfiguration,
    RestApiDataType,
    RestApiInterfaceDescription,
    RestApiInterfaceSelection,
    RestApiServiceCall,
)

__all__ = [
    "ContactFunctionalProfile",
    "ContactFunctionalProfileList",
    "ContactInterface",
    "ContactInterfaceDescription",
    "ContactsDataPointList",
    "GenericDataPointList",
    "GenericFunctionalProfile",
    "GenericFunctionalProfileList",
    "GenericInterface",
    "ModbusDataPoint",
    "ModbusDataPointList",
    "ModbusFunctionalProfile",
    "ModbusFunctionalProfileList",
    "ModbusInterface",
    "AccessProtectionEnabled",
    "BitOrder",
    "MasterFunctionsSupported",
    "MasterFunctionsSupportedList",
    "ModbusAttributes",
    "ModbusBoolean",
    "ModbusDataPointConfiguration",
    "ModbusDataType",
    "ModbusExceptionCode",
    "ModbusInterfaceDescription",
    "ModbusInterfaceSelection",
    "ModbusLayer6Deviation",
    "ModbusRtu",
    "ModbusTcp",
    "RegisterType",
    "TimeSyncBlockNotification",
    "ConfigurationDescription",
    "ConfigurationList",
    "ConfigurationListElement",
    "DeviceFrame",
    "DeviceInformation",
    "InterfaceList",
    "RestApiDataPoint",
    "RestApiDataPointList",
    "RestApiFunctionalProfile",
    "RestApiFunctionalProfileList",
    "RestApiInterface",
    "HeaderEntry",
    "HeaderList",
    "HttpMethod",
    "JmespathMapping",
    "JmespathMappingRecord",
    "ResponseQuery",
    "ResponseQueryType",
    "RestApiAuthenticationMethod",
    "RestApiBasic",
    "RestApiBearer",
    "RestApiDataPointConfiguration",
    "RestApiDataType",
    "RestApiInterfaceDescription",
    "RestApiInterfaceSelection",
    "RestApiServiceCall",
]



from enum import Enum


class SubSetUnits(Enum):

    # Power
    WATTS = 'WATTS',
    KILOWATTS = 'KILOWATTS'
    VOLTS = 'VOLTS',
    AMPERES = 'AMPERES'
    KILOVOLT_AMPERES_REACTIVE = 'KILOVOLT_AMPERES_REACTIVE'
    KILOWATT_HOURS = 'KILOWATT_HOURS'
    KILOVOLT_AMPERES_REACTIVE_HOURS = 'KILOVOLT_AMPERES_REACTIVE_HOURS'

    # Cycle
    HERTZ = 'HERTZ',

    # Temperature
    DEGREES_CELSIUS = 'DEGREES_CELSIUS'

    # Time
    HOURS = 'HOURS'
    MINUTES = 'MINUTES'
    SECONDS = 'SECONDS'

    PERCENT = 'PERCENT'

    NONE = 'NONE'



from typing import Any

from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.context import XmlContext
import time

from sgr_library.api import DeviceInformation, FunctionProfile, DataPointProtocol, DataPoint, ConfigurationParameter, \
    build_configurations_parameters
from sgr_library.api.device_api import BaseSGrInterface
from sgr_library.converters import build_converter
from sgrspecification.generic import DataDirectionProduct
from sgr_library.exceptions import DataPointException, FunctionalProfileException, DataProcessingError, \
    DeviceInformationError, InvalidEndianType
from pymodbus.exceptions import ConnectionException
from aiohttp import ClientError

# from sgr_library.data_classes.ei_modbus import SgrModbusDeviceFrame
# from sgr_library.data_classes.ei_modbus.sgr_modbus_eidevice_frame import SgrModbusDataPointType

from sgrspecification.product import DeviceFrame, ModbusDataPoint, ModbusFunctionalProfile
from sgr_library.auxiliary_functions import get_address, get_endian, get_port, get_slave
from sgrspecification.generic import DataDirectionProduct

from sgr_library.modbus_client import SGrModbusClient
# from auxiliary_functions import find_dp
import asyncio

import logging

from sgr_library.validators import build_validator

logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger(__name__)


class ModBusTCPDataPoint(DataPointProtocol):

    def __init__(self, modbus_api_dp: ModbusDataPoint, modbus_api_fp: ModbusFunctionalProfile,
                 interface: 'SgrModbusInterface'):
        self._dp = modbus_api_dp
        self._fp = modbus_api_fp
        self._interface = interface

    async def write(self, data: Any):
        return await self._interface.setval(self.name()[0], self.name()[1], data)

    async def read(self) -> Any:
        return await self._interface.getval(self.name()[0], self.name()[1])

    def name(self) -> tuple[str, str]:
        return self._fp.functional_profile.functional_profile_name, self._dp.data_point.data_point_name

    def direction(self) -> DataDirectionProduct:
        return self._dp.data_point.data_direction


def build_modbus_tcp_data_point(dp: ModbusDataPoint, fp: ModbusFunctionalProfile,
                                interface: 'SgrModbusInterface') -> DataPoint:
    protocol = ModBusTCPDataPoint(dp, fp, interface)
    converter = build_converter(dp.data_point.unit)
    validator = build_validator(dp.data_point.data_type)
    return DataPoint(protocol, converter, validator)


class ModBusTCPFunctionProfile(FunctionProfile):

    def __init__(self, modbus_api_fp: ModbusFunctionalProfile, interface: 'SgrModbusInterface'):
        self._fp = modbus_api_fp
        self._interface = interface
        dps = [build_modbus_tcp_data_point(dp, self._fp, self._interface) for dp in
               self._fp.data_point_list.data_point_list_element]
        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        return self._fp.functional_profile.functional_profile_name

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SgrModbusInterface(BaseSGrInterface):

    def __init__(self, frame: DeviceFrame) -> None:
        """
        Creates a connection from xml file data.
        Parses the xml file with xsdata library.
        :param xml_file: Name of the xml file to parse
        """
        self.root = frame
        self.ip = get_address(self.root)
        self.port = get_port(self.root)
        self.client = SGrModbusClient(self.ip, self.port)
        self.slave_id = get_slave(self.root)
        self.byte_order = get_endian(self.root)
        self._configuration_params = build_configurations_parameters(frame.configuration_list)
        # A dictionary where we cash the value of the datapoint. With name, value, timestamp and alive_time? ;)
        self.cash_dict = {}
        fps = [ModBusTCPFunctionProfile(profile, self) for profile in
               self.root.interface_list.modbus_interface.functional_profile_list.functional_profile_list_element]
        self._function_profiles = {fp.name(): fp for fp in fps}
        self._device_information = DeviceInformation(
            name=frame.device_name,
            manufacture=frame.manufacturer_name,
            software_revision=frame.device_information.software_revision,
            hardware_revision=frame.device_information.hardware_revision,
            device_category=frame.device_information.device_category,
            is_local=frame.device_information.is_local_control
        )
        self._configuration_parameters = build_configurations_parameters(frame.configuration_list)

    def configuration_parameter(self) -> list[ConfigurationParameter]:
        return self._configuration_parameters

    async def connect(self):
        try:
            await self.client.client.connect()
            logger.info("Connected successfully.")
        except ConnectionException as e:
            logger.error(f"ConnectionException: Failed to connect: {e}")
        except TimeoutError as e:
            logger.error(f"TimeoutError: Connection timed out: {e}")
        except Exception as e:
            logger.exception(f"An unexpected error occurred during the connection: {e}")

    def get_function_profiles(self) -> dict[str, FunctionProfile]:
        return self._function_profiles

    def device_information(self) -> DeviceInformation:
        return self._device_information

    async def getval(self, fp_name, dp_name) -> float:
        """
        Reads datapoint value.

        :dp: The already obtained datapoint object

        2 parameters alternative:
        :param fp_name: The name of the functional profile in which the datapoint resides.
        :param dp_name: The name of the datapoint.

        :returns: The current decoded value in the datapoint register.
        """

        try:
            dp = self.find_dp(self.root, fp_name, dp_name)
        except DataPointException as e:
            logger.error(e)
            return float('nan')
        except FunctionalProfileException as e:
            logger.warning(e)
            return float('nan')

        try:
            address = self.get_address(dp)
            size = self.get_size(dp)
            data_type = self.get_datatype(dp)
            reg_type = self.get_register_type(dp)
            slave_id = self.slave_id
            order = self.byte_order
            answer = await self.client.value_decoder(address, size, data_type, reg_type, slave_id, order)
            return answer
        except ClientError as e:
            logger.exception(e)
            return float('nan')

    # TODO under construction
    async def getval_block(self, fp_name: str, dp_name: str):
        try:
            dp = self.find_dp(self.root, fp_name, dp_name)
        except DataPointException as e:
            logger.error(f"DataPointException: {e}")
            return None
        except FunctionalProfileException as e:
            logger.warning(f"FunctionalProfileException: {e}")
            return None

        try:
            address = self.get_address(dp)
            size = self.get_size(dp)
            data_type = self.get_datatype(dp)
            reg_type = self.get_register_type(dp)
            slave_id = self.slave_id
            order = self.byte_order
            answer = await self.client.mult_value_decoder(address, size, data_type, reg_type, slave_id, order)
            return answer
        except ClientError as e:
            logger.exception(f"ClientError: Failed to retrieve value block {e}")
            return None
        except Exception as e:
            logger.exception(f"An unexpected error occurred: {e}")
            return None

    async def setval(self, fp_name: str, dp_name: str, value: float) -> None:
        try:
            dp = self.find_dp(self.root, fp_name, dp_name)
        except DataPointException as e:
            logger.error(f"DataPointException: {e}")
            return
        except FunctionalProfileException as e:
            logger.warning(f"FunctionalProfileException: {e}")
            return

        try:
            address = self.get_address(dp)
            data_type = self.get_datatype(dp)
            slave_id = self.slave_id
            order = self.byte_order
            await self.client.value_encoder(address, value, data_type, slave_id, order)
            logger.info(f"Value {value} has been set for {dp_name} in {fp_name}")
        except ClientError as e:
            logger.exception(f"ClientError: Failed to set value {e}")
        except ValueError as e:
            logger.error(f"ValueError: Invalid value or datatype {e}")
        except Exception as e:
            logger.exception(f"An unexpected error occurred: {e}")

    def get_device_profile(self):
        try:
            brand_name = self.root.device_information.brand_name
            nominal_power = self.root.device_information.nominal_power
            level_of_operation = self.root.device_information.legible_description

            if None in [brand_name, nominal_power, level_of_operation]:
                raise DeviceInformationError("Incomplete device information")

            logger.info(f"Brand Name: {brand_name}")
            logger.info(f"Nominal Power: {nominal_power}")
            logger.info(f"Level of Operation: {level_of_operation}")

            return self.root.device_information

        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None
        except DeviceInformationError as e:
            logger.error(f"DeviceInformationError: {e}")
            return None

    def get_register_type(self, dp) -> str:
        """
        Returns register type E.g. "HoldRegister"
        :param fp_name: The name of the functional profile
        :param dp_name: The name of the data point.
        :returns: The type of the register
        """
        try:
            register_type = dp.modbus_data_point_configuration.register_type.value
            if register_type is None:
                raise DataProcessingError("Register type is None")
            return register_type
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_datatype(self, dp) -> str:
        try:
            datatype = dp.modbus_data_point_configuration.modbus_data_type.__dict__
            for key in datatype:
                if datatype[key] is not None:
                    return key
            raise DataProcessingError('data_type not available')
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_bit_rank(self, dp):
        try:
            bitrank = dp.modbus_data_point_configuration.bit_rank
            if bitrank is None:
                raise DataProcessingError("Bit rank is None")
            return bitrank
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_address(self, dp):
        try:
            address = dp.modbus_data_point_configuration.address
            if address is None:
                raise DataProcessingError("Address is None")
            return address
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_size(self, dp):
        try:
            size = dp.modbus_data_point_configuration.number_of_registers
            if size is None:
                raise DataProcessingError("Size is None")
            return size
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_multiplicator(self, dp):
        try:
            multiplicator = dp.modbus_attributes.scaling_factor.multiplicator
            if multiplicator is None:
                raise DataProcessingError("Multiplicator is None")
            return multiplicator
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_power_10(self, dp):
        try:
            power_10 = dp.modbus_attributes.scaling_factor.powerof10
            if power_10 is None:
                raise DataProcessingError("Power of 10 is None")
            return power_10
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_unit(self, dp):
        try:
            unit = dp.data_point.unit.value
            if unit is None:
                raise DataProcessingError("Unit is None")
            return unit
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def get_name(self, dp):
        try:
            name = dp.data_point.data_point_name
            if name is None:
                raise DataProcessingError("Name is None")
            return name
        except AttributeError as e:
            logger.error(f"AttributeError: {e}")
            return None

    def find_dp(self, root, fp_name: str, dp_name: str):
        """
        Searches the datapoint in the root element.
        :param root: The root element created with the xsdata parser
        :param fp_name: The name of the funcitonal profile in which the datapoint resides
        :param dp_name: The name of the datapoint
        :returns: The datapoint element found in root, if not, returns None.
        """
        fp = next(filter(lambda x: x.functional_profile.functional_profile_name == fp_name,
                         root.interface_list.modbus_interface.functional_profile_list.functional_profile_list_element),
                  None)
        if fp:
            dp = next(
                filter(lambda x: x.data_point.data_point_name == dp_name, fp.data_point_list.data_point_list_element),
                None)
            if dp:
                return dp
            raise DataPointException(f"Datapoint {dp_name} not found in functional profile {fp_name}.")
        raise FunctionalProfileException(f"Functional profile {fp_name} not found in XML file.")

    # TODO a getval for L1, L2 and L3 at the same time

    # TODO
    def get_dp_attribute(self, datapoint: str, attribute: str):
        """
        Searches for a specific attribute in the datapoint via a key.
        :param attribute"address", "size", "bitrank", "data_type", "register_type", "unit", "multiplicator", "power_of", "name"
        :returns: The chosen attribute.
        """
        # TODO
        ...


async def test_loop():
    while True:
        print('start')
        interface_file = 'abb_terra_01.xml'
        sgr_modbus = SgrModbusInterface(interface_file)
        await sgr_modbus.client.client.connect()
        getval = await sgr_modbus.getval('CurrentAC', 'CurrentACL1')
        print(getval)
        await asyncio.sleep(10)


if __name__ == "__main__":
    starting_time = time.time()
    print('start')
    try:
        asyncio.run(test_loop())
    except KeyboardInterrupt:
        print("done")
